"""
parse the tb files in Wannier90
"""

import os
import re
import numpy as np
from typing import List, Tuple, Dict


def ssrline(file):
    """
    Read a line and split it into a list of strings"""
    line = file.readline().strip()
    return re.split(r"\s+", line)


def parse_vector(file, vector_type, size):
    result = np.zeros((size,), dtype=vector_type)
    for i in range(size):
        result[i] = np.fromstring(file.readline().strip(), sep=" ")
        result[i] = result[i].astype(vector_type)
    return result


def find_Rvectors(Rvectors, R):
    for i, Rvec in enumerate(Rvectors):
        if np.all(Rvec == R):
            return i
    return -1


def parse_tb_file(filename):
    with open(filename, "r") as io:
        header = io.readline().strip()
        a1 = np.fromstring(io.readline().strip(), sep=" ")
        a2 = np.fromstring(io.readline().strip(), sep=" ")
        a3 = np.fromstring(io.readline().strip(), sep=" ")
        # a1, a2, a3 = np.array(list(map(float, [a1, a2, a3])), dtype=np.float64)
        # lattice = np.vstack((a1.reshape(-1, 1), a2.reshape(-1, 1), a3.reshape(-1, 1))).T
        lattice = np.vstack((a1, a2, a3))

        n_wann = int(ssrline(io)[0])
        n_Rvecs = int(ssrline(io)[0])

        Rdegens = []
        line = io.readline().strip()
        while line != "":
            Rdegens += [int(x) for x in line.split()]
            line = io.readline().strip()
        assert len(Rdegens) == n_Rvecs, "Rdegens length does not match n_Rvecs"

        Rvectors = np.zeros((n_Rvecs, 3), dtype=np.int32)
        # H = np.zeros((n_Rvecs,n_wann, n_wann), dtype=np.complex128)
        H = {}
        r_x = np.zeros((n_Rvecs, n_wann, n_wann), dtype=np.complex128)
        r_y = np.zeros((n_Rvecs, n_wann, n_wann), dtype=np.complex128)
        r_z = np.zeros((n_Rvecs, n_wann, n_wann), dtype=np.complex128)
        pos_operator = np.zeros((n_Rvecs, n_wann, n_wann, 3), dtype=np.complex128)

        for iR in range(n_Rvecs):
            # read Rvectors
            Rvectors[iR] = np.fromstring(io.readline().strip(), sep=" ")
            Rvectors[iR] = Rvectors[iR].astype(np.int32)
            R = tuple(Rvectors[iR])
            H[R] = np.zeros((n_wann, n_wann), dtype=np.complex128)
            # read H
            for n in range(n_wann):
                for m in range(n_wann):
                    line = ssrline(io)
                    assert (
                        m == int(line[0]) - 1 and n == int(line[1]) - 1
                    ), "Unexpected indices"
                    reH, imH = float(line[2]), float(line[3])
                    # H[iR][m, n] = reH + 1j * imH
                    H[R][m, n] = (reH + 1j * imH) / 2.0
            io.readline()  # empty line
        # set the onsite term to half
        # np.fill_diagonal(H[(0, 0, 0)], H[(0, 0, 0)].diagonal() / 2.0)
        # print("onsite H from TB: ", H[(0, 0, 0)].diagonal())

        iR0 = find_Rvectors(Rvectors, [0, 0, 0])
        for iR in range(n_Rvecs):
            # read Rvectors
            line = io.readline().strip()
            Rvectors[iR] = np.fromstring(line, sep=" ")
            Rvectors[iR] = Rvectors[iR].astype(np.int32)
            # read r_x, r_y, r_z
            for n in range(n_wann):
                for m in range(n_wann):
                    line = ssrline(io)
                    assert (
                        m == int(line[0]) - 1 and n == int(line[1]) - 1
                    ), f"Unexpected indices in line {line}"
                    pos_operator[iR, m, n, 0] = complex(float(line[2]), float(line[3]))
                    pos_operator[iR, m, n, 1] = complex(float(line[4]), float(line[5]))
                    pos_operator[iR, m, n, 2] = complex(float(line[6]), float(line[7]))
            io.readline()

        # print(
        #    f"Reading tb.dat file: {filename} | Header: {header} | n_wann: {n_wann} | n_Rvecs: {n_Rvecs}"
        # )

        centers = pos_operator[iR0, :, :, :].diagonal(offset=0, axis1=0, axis2=1).T.real

        return {
            "n_wann": n_wann,
            "lattice": lattice,
            "Rvectors": Rvectors,
            "Rdegens": Rdegens,
            "H": H,
            "pos_operator": pos_operator,
            "centers": centers,
            "header": header,
        }


def test_parse_tb_file():
    filename = "cri3_up_tb.dat"
    result = parse_tb_file(filename)
    # rvecs=result["Rvectors"]
    # iR=find_Rvectors(rvecs, [0, 0, 0])
    # for i in range(30):
    #    print(np.real(result["pos_operator"][iR, i, i, :]))
    print(result["centers"])


if __name__ == "__main__":
    test_parse_tb_file()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The abacus wrapper
"""
from pathlib import Path
import os
import numpy as np
from scipy.linalg import eigh
from TB2J.utils import symbol_number_list
from TB2J.myTB import AbstractTB
from TB2J.abacus.abacus_api import read_HR_SR
from TB2J.abacus.orbital_api import parse_abacus_orbital
from TB2J.abacus.stru_api import read_abacus, read_abacus_out


class AbacusWrapper(AbstractTB):
    def __init__(self, HR, SR, Rlist, nbasis, nspin=1):
        self.R2kfactor = -2j * np.pi
        self.is_orthogonal = False
        self.is_siesta = False
        self._name = "Abacus"
        self.HR = HR
        self.SR = SR
        self.Rlist = Rlist
        self.nbasis = nbasis
        self.nspin = nspin
        self.norb = nbasis * nspin
        self._build_Rdict()

    def _build_Rdict(self):
        if hasattr(self, "Rdict"):
            pass
        else:
            self.Rdict = {}
            for iR, R in enumerate(self.Rlist):
                self.Rdict[tuple(R)] = iR

    def get_hamR(self, R):
        return self.HR[self.Rdict[tuple(R)]]

    def gen_ham(self, k, convention=2):
        """
        generate hamiltonian matrix at k point.
        H_k( i, j)=\sum_R H_R(i, j)^phase.
        There are two conventions,
        first:
        phase =e^{ik(R+rj-ri)}. often better used for berry phase.
        second:
        phase= e^{ikR}. We use the first convention here.

        :param k: kpoint
        :param convention: 1 or 2.
        """
        Hk = np.zeros((self.nbasis, self.nbasis), dtype="complex")
        Sk = np.zeros((self.nbasis, self.nbasis), dtype="complex")
        if convention == 2:
            for iR, R in enumerate(self.Rlist):
                phase = np.exp(self.R2kfactor * np.dot(k, R))
                H = self.HR[iR] * phase
                # Hk += H + H.conjugate().T
                Hk += H
                S = self.SR[iR] * phase
                # Sk += S + S.conjugate().T
                Sk += S
        elif convention == 1:
            # TODO: implement the first convention (the r convention)
            raise NotImplementedError("convention 1 is not implemented yet.")
            pass
        else:
            raise ValueError("convention should be either 1 or 2.")
        return Hk, Sk

    def solve(self, k, convention=2):
        Hk, Sk = self.gen_ham(k, convention=convention)
        return eigh(Hk, Sk)

    def HSE_k(self, kpt, convention=2):
        H, S = self.gen_ham(tuple(kpt), convention=convention)
        evals, evecs = eigh(H, S)
        return H, S, evals, evecs

    def HS_and_eigen(self, kpts, convention=2):
        """
        calculate eigens for all kpoints.
        :param kpts: list of k points.
        """
        nk = len(kpts)
        hams = np.zeros((nk, self.nbasis, self.nbasis), dtype=complex)
        evals = np.zeros((nk, self.nbasis), dtype=float)
        evecs = np.zeros((nk, self.nbasis, self.nbasis), dtype=complex)
        for ik, k in enumerate(kpts):
            hams[ik], S, evals[ik], evecs[ik] = self.HSE_k(
                tuple(k), convention=convention
            )
        return hams, None, evals, evecs


class AbacusParser:
    def __init__(self, spin=None, outpath=None, binary=False):
        self.outpath = outpath
        if spin is None:
            self.spin = self.read_spin()
        else:
            self.spin = spin
        self.binary = binary
        # read the information
        self.read_atoms()
        self.efermi = self.read_efermi()
        self.read_basis()

    def read_spin(self):
        with open(str(Path(self.outpath) / "running_scf.log")) as myfile:
            for line in myfile:
                if line.strip().startswith("nspin"):
                    nspin = int(line.strip().split()[-1])
                    if nspin == 1:
                        return "non-polarized"
                    elif nspin == 2:
                        return "collinear"
                    elif nspin == 4:
                        return "noncollinear"
                    else:
                        raise ValueError("nspin should be either 1 or 4.")

    def read_atoms(self):
        path1 = str(Path(self.outpath) / "../STRU")
        path2 = str(Path(self.outpath) / "../Stru")
        if os.path.exists(path1):
            self.atoms = read_abacus(path1)
        elif os.path.exists(path2):
            self.atoms = read_abacus(path2)
        else:
            raise Exception("The STRU or Stru file cannot be found.")
        return self.atoms

    def read_basis(self):
        fname = str(Path(self.outpath) / "Orbital")
        self.basis = parse_abacus_orbital(fname)
        return self.basis

    def read_HSR_collinear(self, binary=None):
        p = Path(self.outpath)
        SR_filename = p / "data-SR-sparse_SPIN0.csr"
        HR_filename = [p / "data-HR-sparse_SPIN0.csr", p / "data-HR-sparse_SPIN1.csr"]
        nbasis, Rlist, HR_up, HR_dn, SR = read_HR_SR(
            nspin=2,
            binary=self.binary,
            HR_fileName=HR_filename,
            SR_fileName=SR_filename,
        )
        return nbasis, Rlist, HR_up, HR_dn, SR

    def Read_HSR_noncollinear(self, binary=None):
        p = Path(self.outpath)
        SR_filename = str(p / "data-SR-sparse_SPIN0.csr")
        HR_filename = str(p / "data-HR-sparse_SPIN0.csr")
        nbasis, Rlist, HR, SR = read_HR_SR(
            nspin=4,
            binary=self.binary,
            HR_fileName=HR_filename,
            SR_fileName=SR_filename,
        )
        return nbasis, Rlist, HR, SR

    def get_models(self):
        if self.spin == "collinear":
            nbasis, Rlist, HR_up, HR_dn, SR = self.read_HSR_collinear()
            model_up = AbacusWrapper(HR_up, SR, Rlist, nbasis, nspin=1)
            model_dn = AbacusWrapper(HR_dn, SR, Rlist, nbasis, nspin=1)
            model_up.efermi = self.efermi
            model_dn.efermi = self.efermi
            model_up.basis, model_dn.basis = self.get_basis()
            model_up.atoms = self.atoms
            model_dn.atoms = self.atoms
            return model_up, model_dn
        elif self.spin == "noncollinear":
            nbasis, Rlist, HR, SR = self.Read_HSR_noncollinear()
            model = AbacusWrapper(HR, SR, Rlist, nbasis, nspin=2)
            model.efermi = self.efermi
            model.basis = self.get_basis()
            model.atoms = self.atoms
            return model

    def read_efermi(self):
        """
        Reading the efermi from the scf log file.
        Search for the line EFERMI = xxxxx eV
        """
        fname = str(Path(self.outpath) / "running_scf.log")
        efermi = None
        with open(fname, "r") as myfile:
            for line in myfile:
                if "EFERMI" in line:
                    efermi = float(line.split()[2])
        if efermi is None:
            raise ValueError(f"EFERMI not found in the {str(fname)}  file.")
        return efermi

    def get_basis(self):
        slist = symbol_number_list(self.atoms)
        if self.spin == "collinear":
            basis_up = []
            basis_dn = []
            for b in self.basis:
                basis_up.append((slist[b.iatom], b.sym, "up"))
                basis_dn.append((slist[b.iatom], b.sym, "down"))
            return basis_up, basis_dn
        elif self.spin == "noncollinear":
            basis = []
            for b in self.basis:
                basis.append((slist[b.iatom], b.sym, "up"))
                basis.append((slist[b.iatom], b.sym, "down"))
            return basis


def test_abacus_wrapper_collinear():
    outpath = "/Users/hexu/projects/TB2J_abacus/abacus-tb2j-master/abacus_example/case_Fe/1_no_soc/OUT.Fe"
    parser = AbacusParser(outpath=outpath, spin=None, binary=False)
    atoms = parser.read_atoms()
    # atoms=parser.read_atoms_out()
    # parser.read_HSR_collinear()
    model_up, model_dn = parser.get_models()
    H, S, E, V = model_up.HSE_k([0, 0, 0])
    # print(H.shape)
    # print(H.diagonal().real)
    # print(model_up.get_HR0().diagonal().real)
    print(parser.efermi)


def test_abacus_wrapper_ncl():
    outpath = "/Users/hexu/projects/TB2J_abacus/abacus-tb2j-master/abacus_example/case_Fe/2_soc/OUT.Fe"

    parser = AbacusParser(outpath=outpath, spin=None, binary=False)
    atoms = parser.read_atoms()
    model = parser.get_models()
    H, S, E, V = model.HSE_k([0, 0, 0])
    print(parser.efermi)


if __name__ == "__main__":
    # test_abacus_wrapper()
    test_abacus_wrapper_ncl()



#!/usr/bin/env python3
from TB2J.abacus.gen_exchange_abacus import gen_exchange_abacus
from TB2J.versioninfo import print_license
import sys
import argparse


def run_abacus2J():
    print_license()
    parser = argparse.ArgumentParser(
        description="abacus2J: Using magnetic force theorem to calculate exchange parameter J from abacus Hamiltonian in the LCAO mode"
    )

    parser.add_argument(
        "--path", help="the path of the abacus calculation", default="./", type=str
    )

    parser.add_argument(
        "--suffix",
        help="the label of the abacus calculation. There should be an output directory called OUT.suffix",
        default="abacus",
        type=str,
    )

    parser.add_argument(
        "--elements",
        help="list of elements to be considered in Heisenberg model.",
        # ,  For each element, a postfixes can be used to specify the orbitals(Only with Siesta backend), eg. Fe_3d, or Fe_3d_4s ",
        default=None,
        type=str,
        nargs="*",
    )
    parser.add_argument(
        "--rcut",
        help="range of R. The default is all the commesurate R to the kmesh",
        default=None,
        type=float,
    )
    parser.add_argument(
        "--efermi", help="Fermi energy in eV. For test only. ", default=None, type=float
    )
    parser.add_argument(
        "--kmesh",
        help="kmesh in the format of kx ky kz. Monkhorst pack. If all the numbers are odd, it is Gamma cenetered. (strongly recommended), Default: 5 5 5",
        type=int,
        nargs="*",
        default=[5, 5, 5],
    )
    parser.add_argument(
        "--emin",
        help="energy minimum below efermi, default -14 eV",
        type=float,
        default=-14.0,
    )

    parser.add_argument(
        "--use_cache",
        help="whether to use disk file for temporary storing wavefunctions and hamiltonian to reduce memory usage. Default: False",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--nz", help="number of integration steps. Default: 50", default=50, type=int
    )

    parser.add_argument(
        "--cutoff",
        help="The minimum of J amplitude to write, (in eV). Default: 1e-7 eV",
        default=1e-7,
        type=float,
    )

    parser.add_argument(
        "--exclude_orbs",
        help="the indices of wannier functions to be excluded from magnetic site. counting start from 0. Default is none.",
        default=[],
        type=int,
        nargs="+",
    )

    parser.add_argument(
        "--np",
        help="number of cpu cores to use in parallel, default: 1",
        default=1,
        type=int,
    )

    parser.add_argument(
        "--description",
        help="add description of the calculatiion to the xml file. Essential information, like the xc functional, U values, magnetic state should be given.",
        type=str,
        default="Calculated with TB2J.",
    )

    parser.add_argument(
        "--orb_decomposition",
        default=False,
        action="store_true",
        help="whether to do orbital decomposition in the non-collinear mode. Default: False.",
    )

    parser.add_argument(
        "--fname",
        default="exchange.xml",
        type=str,
        help="exchange xml file name. default: exchange.xml",
    )

    parser.add_argument(
        "--output_path",
        help="The path of the output directory, default is TB2J_results",
        type=str,
        default="TB2J_results",
    )

    args = parser.parse_args()

    if args.elements is None:
        print("Please input the magnetic elements, e.g. --elements Fe Ni")
        sys.exit()

    include_orbs = {}
    for element in args.elements:
        if "_" in element:
            elem = element.split("_")[0]
            orb = element.split("_")[1:]
            include_orbs[elem] = orb
        else:
            include_orbs[element] = None

    gen_exchange_abacus(
        path=args.path,
        suffix=args.suffix,
        kmesh=args.kmesh,
        magnetic_elements=list(include_orbs.keys()),
        include_orbs=include_orbs,
        Rcut=args.rcut,
        emin=args.emin,
        nz=args.nz,
        description=args.description,
        output_path=args.output_path,
        use_cache=args.use_cache,
        np=args.np,
        exclude_orbs=args.exclude_orbs,
        orb_decomposition=args.orb_decomposition,
    )


if __name__ == "__main__":
    run_abacus2J()



#!/usr/bin/env python3
from TB2J.plot import write_eigen
from TB2J.versioninfo import print_license
import argparse

"""
The script to plot the magnon band structure.
"""


def write_eigen_info():
    print_license()
    parser = argparse.ArgumentParser(
        description="TB2J_eigen.py: Write the eigen values and eigen vectors to file."
    )
    parser.add_argument(
        "--path", default="./", type=str, help="The path of the TB2J_results file"
    )

    parser.add_argument(
        "--qmesh",
        help="qmesh in the format of kx ky kz. Monkhorst pack or Gamma-centered.",
        type=int,
        nargs="*",
        default=[8, 8, 8],
    )

    parser.add_argument(
        "--gamma",
        help="whether shift the qpoint grid to  Gamma-centered. Default: False",
        action="store_true",
        default=True,
    )

    parser.add_argument(
        "--output_fname",
        type=str,
        help="The file name of the output. Default: eigenJq.txt",
        default="eigenJq.txt",
    )

    args = parser.parse_args()

    write_eigen(args.qmesh, args.gamma, output_fname=args.output_fname)


write_eigen_info()



import numpy as np
import os
from itertools import groupby


def write_tom_ucf(cls, fname):
    """
    n1 TODO: what is this
    n2 TODO : what is n2
    id pos_x pos_y pos_z ms damping gyroratio symbol spin_x spin_y spin_z abs(K) K_x K_y K_z
    """
    natom = len(cls.atoms)
    nmatom = len(list(filter(lambda x: x > -1, cls.index_spin)))
    with open(fname, "w") as myfile:
        myfile.write("%s\n" % nmatom)
        myfile.write("%s\n" % nmatom)
        for i in range(natom):
            text = ""
            id_spin = cls.index_spin[i]
            if id_spin > -1:
                pos = cls.atoms.get_scaled_positions()[i]
                ms = np.sqrt(np.sum(np.array(cls.spinat[i]) ** 2))
                spin = np.array(cls.spinat[i]) / ms
                damping = cls.damping[i]
                gyro_ratio = cls.gyro_ratio[i]
                symbol = cls.atoms.get_chemical_symbols()[i]
                if cls.has_uniaxial_anistropy:
                    k1 = cls.k1[i]
                    k1dir = cls.k1dir[i]
                else:
                    k1 = 0.0
                    k1dir = [0.0, 0.0, 1.0]
                text = "{id_spin} {pos_x} {pos_y} {pos_z} {ms} {damping} {gyro_ratio} {symbol} {spin_x} {spin_y} {spin_z} {k1} {kx} {ky} {kz}\n".format(
                    id_spin=id_spin,
                    pos_x=pos[0],
                    pos_y=pos[1],
                    pos_z=pos[2],
                    ms=ms,
                    damping=damping,
                    gyro_ratio=gyro_ratio / 1.76e11,
                    symbol=symbol,
                    spin_x=spin[0],
                    spin_y=spin[1],
                    spin_z=spin[2],
                    k1=k1,
                    kx=k1dir[0],
                    ky=k1dir[1],
                    kz=k1dir[2],
                )
                myfile.write(text)


def write_tom_exch(cls, fname):
    """
    write exchange_J to exch file.
    """
    # TODO convert DMI to matrix form and

    # prepare
    maxInt = 0
    if cls.has_exchange:
        for key, group in groupby(cls.exchange_Jdict.keys(), lambda x: (x[1], x[2])):
            l = len(list(group))
            if l > maxInt:
                maxInt = l

    with open(fname, "w") as myfile:
        myfile.write(
            """exchange:
{
FourSpin = FALSE;
Scale = [ 1.0 , 1.0 , 1.0 ];
MaxInteractions = %s;
TruncateExchange = FALSE;
};

        """
            % maxInt
        )
        if cls.has_exchange:
            for key, group in groupby(
                cls.exchange_Jdict.keys(), lambda x: (x[1], x[2])
            ):
                group = list(group)
                myfile.write("exchange_{}_{}:\n".format(key[0], key[1]))
                myfile.write("{ \n")
                myfile.write("Num_Interactions = {};\n".format(len(list(group))))
                myfile.write('units = "eV";\n')
                for i, k in enumerate(group):
                    myfile.write(
                        "UnitCell{} = [{:.5e} , {:.5e} , {:.5e} ];\n".format(
                            i + 1, k[0][0], k[0][1], k[0][2]
                        )
                    )
                    myfile.write(
                        "J{}_{} = [{:.8e} , {:.8e}, {:.8e}];\n".format(
                            i + 1, 1, cls.exchange_Jdict[k], 0.0, 0.0
                        )
                    )
                    myfile.write(
                        "J{}_{} = [{:.8e} , {:.8e}, {:.8e}];\n".format(
                            i + 1, 2, 0.0, cls.exchange_Jdict[k], 0.0
                        )
                    )
                    myfile.write(
                        "J{}_{} = [{:.8e} , {:.8e}, {:.8e}];\n".format(
                            i + 1, 3, 0.0, 0.0, cls.exchange_Jdict[k]
                        )
                    )
                myfile.write("};\n")


# Tom's ASD code


def write_tom_format(cls, path="TB2J_results/TomASD", prefix="exchange"):
    if not os.path.exists(path):
        os.makedirs(path)
    exch_fname = os.path.join(path, "%s.exch" % prefix)
    ucf_fname = os.path.join(path, "%s.ucf" % prefix)
    write_tom_ucf(cls, ucf_fname)
    write_tom_exch(cls, exch_fname)



import os
import numpy as np
from ase.atoms import Atoms
from TB2J.utils import symbol_number
from collections import defaultdict
from scipy.linalg import eigh
from TB2J.myTB import AbstractTB
from TB2J.mathutils import Lowdin


class SislWrapper(AbstractTB):
    def __init__(self, sisl_hamiltonian, geom=None, spin=None):
        self.is_siesta = False
        self.is_orthogonal = False
        self.ham = sisl_hamiltonian
        # k2Rfactor : H(k) = \int_R H(R) * e^(k2Rfactor * k.R)
        self.R2kfactor = 2.0j * np.pi  #
        if spin == "up":
            spin = 0
        elif spin == "down":
            spin = 1
        if spin not in [None, 0, 1, "merge", "fakemerge"]:
            raise ValueError("spin should be None/0/1, but is %s" % spin)
        self.spin = spin
        self.orbs = []
        self.orb_dict = defaultdict(lambda: [])
        if geom is None:
            g = self.ham._geometry
        else:
            g = geom
        _atoms = geom._atoms
        atomic_numbers = []
        self.positions = g.xyz
        self.cell = np.array(g.sc.cell)
        for ia, a in enumerate(_atoms):
            atomic_numbers.append(a.Z % 200)
        self.atoms = Atoms(
            numbers=atomic_numbers, cell=self.cell, positions=self.positions
        )
        sdict = list(symbol_number(self.atoms).keys())
        if self.ham.spin.is_colinear and (self.spin in [0, 1]):
            for ia, a in enumerate(_atoms):
                symnum = sdict[ia]
                try:
                    orb_names = [f"{symnum}|{x.name()}|up" for x in a.orbital]
                except:
                    orb_names = [f"{symnum}|{x.name()}|up" for x in a.orbitals]
                self.orbs += orb_names
                self.orb_dict[ia] += orb_names
            self.norb = len(self.orbs)
            self.nbasis = self.norb
        elif (
            self.ham.spin.is_spinorbit
            or self.ham.spin.is_noncolinear
            or self.spin == "merge"
            or self.spin == "fakemerge"
        ):
            for ia, a in enumerate(_atoms):
                symnum = sdict[ia]
                orb_names = []
                for x in a.orbitals:
                    for spin in {"up", "down"}:
                        orb_names.append(f"{symnum}|{x.name()}|{spin}")
                self.orbs += orb_names
                self.orb_dict[ia] += orb_names
            self.norb = len(self.orbs) // 2
            self.nbasis = len(self.orbs)
        else:
            raise ValueError("The hamiltonian should be either spin-orbit or colinear")
        self._name = "SIESTA"

    def view_info(self):
        print(self.orb_dict)
        print(self.atoms)

    def solve(self, k, convention=2):
        if convention == 1:
            gauge = "r"
        elif convention == 2:
            gauge = "R"
        if self.spin in [0, 1]:
            evals, evecs = self.ham.eigh(
                k=k, spin=self.spin, eigvals_only=False, gauge=gauge
            )
        elif self.spin is None:
            evals, evecs = self.ham.eigh(k=k, eigvals_only=False, gauge=gauge)
        elif self.spin == "merge":
            evals0, evecs0 = self.ham.eigh(k=k, spin=0, eigvals_only=False, gauge=gauge)
            evals1, evecs1 = self.ham.eigh(k=k, spin=1, eigvals_only=False, gauge=gauge)
            evals = np.zeros(self.nbasis, dtype=float)
            evecs = np.zeros((self.nbasis, self.nbasis), dtype=complex)
            evals[::2] = evals0
            evals[1::2] = evals1
            evecs[::2, ::2] = evecs0
            evecs[1::2, 1::2] = evecs1

        elif self.spin == "fakemerge":
            evals0, evecs0 = self.ham.eigh(k=k, eigvals_only=False, gauge=gauge)
            evals1, evecs1 = self.ham.eigh(k=k, eigvals_only=False, gauge=gauge)
            evals = np.zeros(self.nbasis, dtype=float)
            evecs = np.zeros((self.nbasis, self.nbasis), dtype=complex)
            evals[::2] = evals0
            evals[1::2] = evals1
            evecs[::2, ::2] = evecs0
            evecs[1::2, 1::2] = evecs1
        return evals, evecs

    def Hk(self, k, convention=2):
        if convention == 1:
            gauge = "r"
        elif convention == 2:
            gauge = "R"
        if self.spin is None:
            H = self.ham.Hk(k, gauge=gauge, format="dense")
        elif self.spin in [0, 1]:
            H = self.ham.Hk(k, spin=self.spin, gauge=gauge, format="dense")
        elif self.spin == "merge":
            H = np.zeros((self.nbasis, self.nbasis), dtype="complex")
            H[::2, ::2] = self.ham.Hk(k, spin=0, gauge=gauge, format="dense")
            H[1::2, 1::2] = self.ham.Hk(k, spin=1, gauge=gauge, format="dense")
        elif self.spin == "fakemerge":
            H = np.zeros((self.nbasis, self.nbasis), dtype="complex")
            H[::2, ::2] = self.ham.Hk(k, gauge=gauge, format="dense")
            H[1::2, 1::2] = self.ham.Hk(k, gauge=gauge, format="dense")
        return H

    def eigen(self, k, convention=2):
        return self.solve(k)

    def gen_ham(self, k, convention=2):
        return self.Hk(k, convention=convention)

    def Sk(self, k, convention=2):
        if convention == 1:
            pass
        elif convention == 2:
            pass
        S0 = self.ham.Sk(k, gauge="R", format="dense")
        if self.spin is None:
            S = S0
        elif self.spin in [0, 1]:
            S = S0
        elif self.spin == "merge" or self.spin == "fakemerge":
            S = np.zeros((self.nbasis, self.nbasis), dtype="complex")
            S[::2, ::2] = S0
            S[1::2, 1::2] = S0
        return S

    def solve_all(self, kpts, orth=False):
        evals = []
        evecs = []
        for ik, k in enumerate(kpts):
            if orth:
                S = self.Sk(k)
                Smh = Lowdin(S)
                H = self.gen_ham(k)
                Horth = Smh.T.conj() @ H @ Smh
                evalue, evec = eigh(Horth)
            else:
                evalue, evec = self.solve(k)
            evals.append(evalue)
            evecs.append(evec)
        return np.array(evals, dtype=float), np.array(evecs, dtype=complex, order="C")

    def HSE_k(self, k, convention=2):
        Hk = self.Hk(k, convention=convention)
        Sk = self.Sk(k, convention=convention)
        evalue, evec = eigh(Hk, Sk)
        # evalue, evec = self.solve(k, convention=convention)
        return Hk, Sk, evalue, evec

    def HS_and_eigen(self, kpts, convention=2):
        nkpts = len(kpts)
        evals = np.zeros((nkpts, self.nbasis), dtype=float)
        self.nkpts = nkpts
        if not self._use_cache:
            evecs = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
            H = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
            S = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
        else:
            self._prepare_cache()

        for ik, k in enumerate(kpts):
            if self._use_cache:
                self.evecs = np.memmap(
                    os.path.join(self.cache_path, "evecs.dat"),
                    mode="w+",
                    shape=(nkpts, self.nbasis, self.nbasis),
                    dtype=complex,
                )
                self.H = np.memmap(
                    os.path.join(self.cache_path, "H.dat"),
                    mode="w+",
                    shape=(nkpts, self.nbasis, self.nbasis),
                    dtype=complex,
                )
                self.S = np.memmap(
                    os.path.join(self.cache_path, "H.dat"),
                    mode="w+",
                    shape=(nkpts, self.nbasis, self.nbasis),
                    dtype=complex,
                )
            Hk = self.Hk(k, convention=convention)
            Sk = self.Sk(k, convention=convention)
            evalue, evec = eigh(Hk, Sk)
            self.H[ik] = Hk
            self.S[ik] = Sk
            self.evals[ik] = evalue
            self.evecs[ik] = evec
            if self._use_cache:
                del self.evecs
                del self.H
                del self.S
        return self.H, self.S, self.evals, self.evecs

    def _prepare_cache(self, path="./TB2J_results/cache"):
        if not os.path.exists(path):
            os.makedirs(path)
        else:
            os.remove(path)
        self.cache_path = path

    # def get_HSE_cached(self, kpt, convention=2):
    #     kpt = tuple(kpt)
    #     kname = f"{kpt[0]:9.5f}_{kpt[1]:9.5f}_{kpt[2]:9.5f}"
    #     if kname in self.cache_dict:
    #         path = self.cache_dict[kname]
    #         Hk = np.memmap(os.path.join(path, 'H.dat'),
    #                        dtype='complex',
    #                        mode='r',
    #                        shape=(self.nbasis, self.nbasis))
    #         Sk = np.memmap(os.path.join(path, 'H.dat'),
    #                        dtype='complex',
    #                        mode='r',
    #                        shape=(self.nbasis, self.nbasis))
    #         evalue = np.memmap(os.path.join(path, 'evalue.dat'),
    #                            dtype='float',
    #                            mode='r',
    #                            shape=(self.nbasis))
    #         evec = np.memmap(os.path.join(path, 'evec.dat'),
    #                          dtype='complex',
    #                          mode='r',
    #                          shape=(self.nbasis, self.nbasis))
    #     else:
    #         Hk, Sk, evalue, evec = self.get_HSE(kpt, convention=convention)
    #         path = os.path.join(self.cache_path, )
    #         self.cache_dict[tuple(kpt)] = path
    #         fpH = np.memmap(os.path.join(path, 'H.dat'))
    #     return Hk, Sk, evalue, evec

    def get_HSE(self, kpt, convention=2):
        Hk = self.Hk(kpt, convention=convention)
        Sk = self.Sk(kpt, convention=convention)
        evalue, evec = self.solve(kpt, convention=convention)
        return Hk, Sk, evalue, evec

    def get_fermi_level(self):
        return 0.0


class SislWFSXWrapper(SislWrapper):
    """Wrapper for retrieving eigenvalues and eigenvectors from siesta WFSX file

    Parameters
    ----------
    geom : sisl.Geometry
        the geometry containing information about atomic positions and orbitals
    wfsx_sile: sisl.io.siesta.wfsxSileSiesta
        file reader for WFSX file
    spin : sisl.physics.Spin
        spin object carrying information about spin configurations and spin component.
    ispin : None or int
        index of spin channel to be considered. Only takes effect for collinear spin calculations (UP: 0, DOWN: 1).
        (default: None)
    shift_fermi: None or float
        energy shift to be applied to all energies. If `None` no shift is applied. (default: None)
    """

    def __init__(
        self, sisl_hamiltonian, geom, wfsx_sile, spin, ispin=None, shift_fermi=None
    ):
        # super().__init__(geom, spin=spin, ispin=ispin, shift_fermi=shift_fermi)
        super().__init__(sisl_hamiltonian, geom=geom, spin=spin)
        self.geom = geom
        self.wfsx_sile = wfsx_sile
        self.read_all()

    def read_all(self):
        """Read all wavefunctions, eigenvalues, and k-points from WFSX file."""
        evals = []
        evecs = []
        wfsx_kpts = []

        def change_gauge(k, evec):
            """Change the eigenvector gauge"""
            phase = np.dot(
                self.geom.xyz[self.geom.o2a(np.arange(self.geom.no)), :],
                np.dot(k, self.geom.rcell),
            )
            if self.spin.has_noncolinear:
                # for NC/SOC we have a 2x2 spin-box per orbital
                phase = np.repeat(phase, 2)
            # r -> R gauge tranformation.
            return evec * np.exp(1j * phase).reshape(1, -1)

        # Read wavefunctions and eigenvalues
        for wfc in self.wfsx_sile.yield_eigenstate(parent=self.geom):
            wfsx_kpts.append(wfc.info["k"])
            evals.append(wfc.c)
            # To get the same eigvecs as eigh returns we need to transpose the
            # array and change the gauge from 'r' to 'R'
            evecs.append(change_gauge(wfc.info["k"], wfc.state).T)

        # If any k-point occurs more than once in the WaveFuncKPoints block,
        # we discard the duplicates
        is_duplicate = self._is_duplicate(wfsx_kpts)
        self.wfsx_kpts = wfsx_kpts[~is_duplicate]
        self.evals = np.array(evals, dtype=float)[~is_duplicate]
        if self.shift_fermi is not None:
            self.evals += self.shift_fermi
        self.evecs = np.array(evecs, dtype=np.complex64, order="C")[~is_duplicate]

    def _is_duplicate(self, array):
        # TODO: Move into utils
        # Find all matches (i,j): array[i] == array[j]
        matches = np.all(np.isclose(array[None, :], array[:, None]), axis=-1)
        # Remove double counting of matches: (i,j) and (j,i)
        # Also, remove matches of elements with themselves: (i,i)
        matches = np.triu(matches, 1)

        # Finally determine elements which are duplicates
        return np.any(matches, axis=0)

    def find_all(self, kpts):
        """Find the correct eigenvectors and eigenvalues and sort them
        to match the order in kpts.

        Parameters
        ----------
        kpts : list of float (3,) or (nk, 3)
            list of k points

        Returns
        -------
        evals : list of float (nk, n)
            list of eigenvalues for every requested k point
        evecs :
            list of eiegnvector for every requested k point
        """
        kpts = np.asarray(kpts)
        sort_idx = np.where(
            np.all(np.isclose(self.wfsx_kpts[None, :], kpts[:, None]), axis=-1)
        )[1]
        if len(sort_idx) < len(kpts):
            # k-point not found
            raise ValueError(
                f"{self.__class__.__name__}._read_all unable to find at least one "
                "required k point in '{self.wfsx_sile.file}'. Please, ensure that "
                "all k points listed below are included:\n"
                + "\n".join([str(k) for k in kpts])
            )
        if not np.all(np.isclose(self.wfsx_kpts[sort_idx], kpts)):
            # raise ValueError(       wfsx_kpts = np.asarray(wfsx_kpts)
            pass


# def test():
#    fdf = sisl.get_sile('/home/hexu/projects/learn_siesta/SMO_Wannier/siesta.fdf')
#    H = fdf.read_hamiltonian(order='nc',dim=2)
#    print(H._spin._is_polarized)
#    print(H.__dict__.keys())
#    print(H._geometry.__dict__.keys())
#    print(H._geometry.xyz)
#    print(H._geometry.sc)
#    H._geometry.sc.cell
#    orb=H._geometry._atoms[0].orbital[0]
#    print(orb.name())
#    s=SislWrapper(H)
#    s.view_info()
#
# import sisl
# fdf = sisl.get_sile('/home/hexu/projects/learn_siesta/SMO_Wannier/siesta.fdf')
# H = fdf.read_hamiltonian(order='nc',dim=2)
# if __name__=='__main__':
#    test()



import numpy as np
from scipy.linalg import eigh, inv


def eigen_to_G(evals, evecs, efermi, energy):
    """calculate green's function from eigenvalue/eigenvector for energy(e-ef): G(e-ef).
    :param evals:  eigen values
    :param evecs:  eigen vectors
    :param efermi: fermi energy
    :param energy: energy
    :returns: Green's function G,
    :rtype:  Matrix with same shape of the Hamiltonian (and eigenvector)
    """
    G = evecs.dot(np.diag(1.0 / (-evals + (energy + efermi)))).dot(evecs.conj().T)
    return G


def green_H(H, energy, S=np.eye(3)):
    return inv(S * energy - H)


def green_H_eig(H, energy):
    evals, evecs = eigh(H)
    return eigen_to_G(evals, evecs, 0.0, energy)


def test_eigh():
    H = np.random.random((3, 3))
    H = H + H.T.conj()
    evals, evecs = eigh(H)
    # print(f"VT@V: {evecs.T.conj()@evecs}")
    green_H(H, 1)
    green_H_eig(H, 1)
    print(f"H: {H}")

    S = np.random.random((3, 3)) + np.random.random((3, 3)) * 1j
    S = S + S.T.conj()
    S = np.eye(3) * 0.4 + S
    evals, evecs = eigh(H, S)

    print(f"VT@V: {evecs.T.conj()@evecs}")
    print(f"VT@S@V: {evecs.T.conj()@S@evecs}")  # I
    print(f"V@S@VT: {evecs@S@evecs.T.conj()}")  # Not I
    print(f"S@VT@evals@V: {S@evecs.T.conj()@np.diag(evals)@evecs}")
    print(f"V@evals@VT: {evecs@np.diag(evals)@evecs.T.conj()}")
    print(f"VT@evals@V: {evecs.T.conj()@np.diag(evals)@evecs}")

    G1 = green_H(H, 0.3, S=S)
    # print("G1=", G1)

    evals, evecs = eigh(H, S)
    G2 = eigen_to_G(evals, evecs, 0.3, 0)
    G2 = green_H(H, 0.3, S=S)
    print(f"G1-G2={G1-G2}")


test_eigh()



#!/usr/bin/env python3
import argparse
from TB2J.rotate_atoms import rotate_atom_xyz, rotate_xyz, check_ftype


def main():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument(
        "fname", help="name of the file containing the atomic structure", type=str
    )
    parser.add_argument(
        "--ftype",
        help="type of the output files, e.g.  xyz. Please use the format which contains the full cell matrix. (e.g. .cif file should not be used) ",
        default="xyz",
        type=str,
    )

    args = parser.parse_args()
    rotate_xyz(args.fname, ftype=args.ftype)


if __name__ == "__main__":
    main()



import os
import copy
from TB2J.myTB import MyTB, merge_tbmodels_spin
from TB2J.exchange_pert import ExchangePert
from TB2J.utils import read_basis, auto_assign_basis_name
from ase.io import read


class PolyTB:
    def __init__(self, ref_model, m0, m1, m2):
        self.ref_model = ref_model
        self.m0 = m0
        self.m1 = m1
        self.m2 = m2

    def gen_model(self, amp):
        m = copy.deepcopy(self.ref_model)
        for R in m.data:
            m.data[R] = (
                self.m0.data[R]
                + self.m1.data[R] * amp
                + self.m2.data[R] * (amp * amp * 0.5)
            )
        return m


def gen_exchange_Oiju(
    path,
    amp,
    colinear=True,
    poscar="POSCAR",
    prefix_up="wannier90.up",
    prefix_dn="wannier90.dn",
    prefix_SOC="wannier90",
    d0_up_ncfile="m0_up.nc",
    d0_dn_ncfile="m0_dn.nc",
    dHdx_up_ncfile="m1_up.nc",
    dHdx_dn_ncfile="m1_dn.nc",
    d2_up_ncfile="m2_up.nc",
    d2_dn_ncfile="m2_dn.nc",
    min_hopping_norm=1e-6,
    max_distance=None,
    efermi=3.0,
    magnetic_elements=[],
    kmesh=[5, 5, 5],
    emin=-12.0,
    emax=0.0,
    exclude_orbs=[],
    Rmesh=[1, 1, 1],
    description="",
):
    atoms = read(os.path.join(path, poscar))
    basis_fname = os.path.join(path, "basisb.txt")
    if colinear:
        tbmodel_up = MyTB.read_from_wannier_dir(
            path=path, prefix=prefix_up, poscar=poscar, nls=False
        )
        tbmodel_dn = MyTB.read_from_wannier_dir(
            path=path, prefix=prefix_dn, poscar=poscar, nls=False
        )
        tbmodel = merge_tbmodels_spin(tbmodel_up, tbmodel_dn)

        dHdx_up = MyTB.load_MyTB(dHdx_up_ncfile)
        dHdx_dn = MyTB.load_MyTB(dHdx_dn_ncfile)
        dHdx = merge_tbmodels_spin(dHdx_up, dHdx_dn)

        d0_up = MyTB.load_MyTB(d0_up_ncfile)
        d0_dn = MyTB.load_MyTB(d0_dn_ncfile)
        d0 = merge_tbmodels_spin(d0_up, d0_dn)

        d2_up = MyTB.load_MyTB(d2_up_ncfile)
        d2_dn = MyTB.load_MyTB(d2_dn_ncfile)
        d2 = merge_tbmodels_spin(d2_up, d2_dn)

        tb = PolyTB(ref_model=tbmodel, m0=d0, m1=dHdx, m2=d2)

        tbmodel = tb.gen_model(amp)

        if os.path.exists(basis_fname):
            basis = read_basis(basis_fname)
        else:
            basis, _ = auto_assign_basis_name(tbmodel.xred, atoms)
        exchange = ExchangePert(
            tbmodels=tbmodel,
            atoms=atoms,
            basis=basis,
            efermi=efermi,
            magnetic_elements=magnetic_elements,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            exclude_orbs=exclude_orbs,
            Rmesh=Rmesh,
            description=description,
        )
        exchange.set_dHdx(dHdx)
        exchange.run("DJ_amp%.3f" % amp)


if __name__ == "__main__":
    gen_exchange_Oiju(
        path="/media/hexu/Backup/materials/SOC/SrMnO3_FM/U3_SrMnO3_slater0.02",
        amp=0.02,
        colinear=True,
        poscar="POSCAR",
        prefix_up="wannier90.up",
        prefix_dn="wannier90.dn",
        prefix_SOC="wannier90",
        dHdx_up_ncfile="m1_up.nc",
        dHdx_dn_ncfile="m1_dn.nc",
        min_hopping_norm=1e-4,
        max_distance=None,
        efermi=3.95,
        magnetic_elements=["Mn"],
        kmesh=[4, 4, 4],
        emin=-12.0,
        emax=0.0,
        exclude_orbs=[],
        Rmesh=[1, 1, 1],
        description="",
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The main function to compute exchange interaction from abacus data
"""

import os
from pathlib import Path
from TB2J.abacus.abacus_wrapper import AbacusParser
from TB2J.exchange import ExchangeNCL, ExchangeCL
from TB2J.exchangeCL2 import ExchangeCL2


def gen_exchange_abacus(
    path,
    suffix="Abacus",
    binary=False,
    magnetic_elements=[],
    include_orbs=[],
    kmesh=[7, 7, 7],
    emin=-13.0,
    emax=0.00,
    nz=100,
    exclude_orbs=[],
    Rcut=None,
    use_cache=False,
    np=1,
    output_path="TB2J_results",
    orb_decomposition=False,
    description=None,
):
    outpath = Path(path) / f"OUT.{suffix}"

    if not os.path.exists(outpath):
        raise ValueError(
            f"The path {outpath} does not exist. Please check if the path and the suffix is correct"
        )
    parser = AbacusParser(outpath=outpath, spin=None, binary=binary)
    spin = parser.read_spin()
    if spin == "collinear":
        tbmodel_up, tbmodel_dn = parser.get_models()
        efermi = parser.read_efermi()
        print("Starting to calculate exchange.")
        description = f""" Input from collinear Abacus data.
data directory: {outpath}
\n"""
        exchange = ExchangeCL2(
            tbmodels=(tbmodel_up, tbmodel_dn),
            atoms=tbmodel_up.atoms,
            basis=tbmodel_up.basis,
            efermi=efermi,
            magnetic_elements=magnetic_elements,
            include_orbs=include_orbs,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            nz=nz,
            exclude_orbs=exclude_orbs,
            Rcut=Rcut,
            np=np,
            use_cache=use_cache,
            output_path=output_path,
            description=description,
        )
        exchange.run(path=output_path)
        print("\n")
        print(f"All calculation finsihed. The results are in {output_path} directory.")
    else:
        tbmodel = parser.get_models()
        print("Starting to calculate exchange.")
        description = f""" Input from non-collinear Abacus data.
data directory: {outpath}
\n"""
        exchange = ExchangeNCL(
            tbmodels=tbmodel,
            atoms=tbmodel.atoms,
            basis=tbmodel.basis,
            efermi=tbmodel.efermi,
            magnetic_elements=magnetic_elements,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            nz=nz,
            exclude_orbs=exclude_orbs,
            Rcut=Rcut,
            np=np,
            use_cache=use_cache,
            description=description,
        )
        exchange.run()
        print("\n")
        print("All calculation finsihed. The results are in TB2J_results directory.")


if __name__ == "__main__":
    gen_exchange_abacus(
        path="/Users/hexu/projects/TB2J_abacus/abacus-tb2j-master/abacus_example/case_Fe/2_soc",
        suffix="Fe",
        magnetic_elements=["Fe"],
        nz=50,
        Rcut=8,
        kmesh=[7, 7, 7],
    )



from collections import defaultdict, OrderedDict
import os
import numpy as np
from TB2J.green import TBGreen
from TB2J.pauli import pauli_block_all, pauli_block_sigma_norm, pauli_mat
from TB2J.utils import symbol_number, kmesh_to_R
from TB2J.io_exchange import SpinIO
from tqdm import tqdm
from TB2J.external import p_map
from TB2J.contour import Contour
from TB2J.utils import simpson_nonuniform, trapezoidal_nonuniform, split_symbol_number
from TB2J.orbmap import map_orbs_matrix
import pickle


class ExchangeParams:
    def __init__(
        self,
        efermi,
        basis=None,
        magnetic_elements=[],
        include_orbs={},
        kmesh=[4, 4, 4],
        emin=-15,  # integration lower bound, relative to fermi energy
        # integration upper bound. Should be 0 (fermi energy). But DFT codes define Fermi energy in various ways.
        emax=0.05,
        nz=100,
        # the delta in the (i delta) in green's function to prevent divergence
        exclude_orbs=[],  #
        ne=None,  # number of electrons in Wannier function.
        Rcut=None,  # Rcut.
        use_cache=False,
        np=1,
        description="",
        write_density_matrix=False,
        orb_decomposition=False,
        output_path="TB2J_results",
    ):
        self.efermi = efermi
        self.emin = emin
        self.emax = emax
        self.nz = nz
        self.Rcut = Rcut
        self.basis = basis
        self.magnetic_elements = magnetic_elements
        self.include_orbs = include_orbs

        self.exclude_orbs = exclude_orbs
        self.ne = ne
        self._use_cache = use_cache
        self.np = np
        self._kmesh = kmesh
        self.orb_decomposition = orb_decomposition
        self.write_density_matrix = write_density_matrix
        self.description = description
        self.output_path = output_path


class Exchange(ExchangeParams):
    def __init__(
        self,
        tbmodels,
        atoms,
        efermi,
        basis=None,
        magnetic_elements=[],
        include_orbs={},
        kmesh=[4, 4, 4],
        emin=-15,  # integration lower bound, relative to fermi energy
        # integration upper bound. Should be 0 (fermi energy). But DFT codes define Fermi energy in various ways.
        emax=0.05,
        nz=100,
        # the delta in the (i delta) in green's function to prevent divergence
        exclude_orbs=[],  #
        ne=None,  # number of electrons in Wannier function.
        Rcut=None,  # Rcut.
        use_cache=False,
        np=1,
        description="",
        write_density_matrix=False,
        output_path="TB2J_results",
        orb_decomposition=False,
    ):
        self.atoms = atoms
        super().__init__(
            efermi=efermi,
            basis=basis,
            magnetic_elements=magnetic_elements,
            include_orbs=include_orbs,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            nz=nz,
            exclude_orbs=exclude_orbs,
            ne=ne,
            Rcut=Rcut,
            use_cache=use_cache,
            np=np,
            description=description,
            write_density_matrix=write_density_matrix,
            orb_decomposition=orb_decomposition,
            output_path=output_path,
        )
        self._prepare_kmesh(self._kmesh)
        self._prepare_Rlist()
        self.set_tbmodels(tbmodels)
        self._adjust_emin()
        self._prepare_elist(method="legendre")
        self._prepare_basis()
        self._prepare_orb_dict()
        self._prepare_distance()

        # whether to calculate J and DMI with NJt method.
        self.calc_NJt = True
        # self._prepare_NijR()
        self.Ddict_NJT = None
        self.Jdict_NJT = None
        self._is_collinear = True
        self.has_elistc = False
        self._clean_tbmodels()

    def _prepare_Jorb_file(self):
        os.makedirs(self.output_path, exist_ok=True)
        self.orbpath = os.path.join(self.output_path, "OrbResolve")
        os.makedirs(self.orbpath, exist_ok=True)

    def _adjust_emin(self):
        self.emin = self.G.find_energy_ingap(rbound=self.efermi - 5.0) - self.efermi
        print(f"A gap is found at {self.emin}, set emin to it.")

    def set_tbmodels(self, tbmodels):
        pass

    def _clean_tbmodels(self):
        del self.tbmodel
        del self.G.tbmodel

    def _prepare_kmesh(self, kmesh):
        for k in kmesh:
            self.kmesh = list(map(lambda x: x // 2 * 2 + 1, kmesh))

    def _prepare_elist(self, method="legendre"):
        """
        prepare list of energy for integration.
        The path has three segments:
         emin --1-> emin + 1j*height --2-> emax+1j*height --3-> emax
        """
        self.contour = Contour(self.emin, self.emax)
        # if method.lower() == "rectangle":
        #    self.contour.build_path_rectangle(
        #        height=self.height, nz1=self.nz1, nz2=self.nz2, nz3=self.nz3
        #    )
        if method.lower() == "semicircle":
            self.contour.build_path_semicircle(npoints=self.nz, endpoint=True)
        elif method.lower() == "legendre":
            self.contour.build_path_legendre(npoints=self.nz, endpoint=True)
        else:
            raise ValueError(f"The path cannot be of type {method}.")

    def _prepare_Rlist(self):
        """
        prepare R list for J(i, j, R)
        [-Rx, Rx] * [-Ry, Ry] * [-Rz, Rz]
        """
        self.Rlist = kmesh_to_R(self.kmesh)

    def _prepare_basis(self):
        if self.basis is None:
            pass

    def _prepare_orb_dict(self):
        """
        generate self.ind_mag_atoms and self.orb_dict
        """
        # adict: dictionary of {'Fe': ['dxy', 'dyz', ...], ...}
        adict = OrderedDict()
        # orb_dict: {ind_atom:[ind_orb,1,2], ...}
        self.orb_dict = {}
        # labels: {0:{dxy, ...}}
        self.labels = {}
        # magnetic atoms index
        self.ind_mag_atoms = []

        sdict = symbol_number(self.atoms)

        for i, base in enumerate(self.basis):
            if i not in self.exclude_orbs:
                # e.g. Fe2, dxy, _, _
                if isinstance(base, str):
                    atom_sym, orb_sym = base.split("|")[:2]
                else:
                    atom_sym, orb_sym = base[:2]

                if atom_sym in adict:
                    adict[atom_sym].append(orb_sym)
                else:
                    adict[atom_sym] = [orb_sym]
                iatom = sdict[atom_sym]
                if iatom not in self.orb_dict:
                    self.orb_dict[iatom] = [i]
                    self.labels[iatom] = [orb_sym]
                else:
                    self.orb_dict[iatom] += [i]
                    self.labels[iatom] += [orb_sym]

        # index of magnetic atoms
        symbols = self.atoms.get_chemical_symbols()
        tags = self.atoms.get_tags()
        for i, (sym, tag) in enumerate(zip(symbols, tags)):
            if sym in self.magnetic_elements or f"{sym}{tag}" in self.magnetic_elements:
                self.ind_mag_atoms.append(i)

        # sanity check to see if some magnetic atom has no orbital.
        for iatom in self.ind_mag_atoms:
            if iatom not in self.orb_dict:
                raise ValueError(
                    f"""Cannot find any orbital for atom {iatom}, which is supposed to be magnetic. Please check the Wannier functions."""
                )
        if not self._is_collinear:
            for iatom, orb in self.orb_dict.items():
                nsorb = len(self.orb_dict[iatom])
                if nsorb % 2 != 0:
                    raise ValueError(
                        f"""The number of spin-orbitals for atom {iatom} is not even,
{nsorb} spin-orbitals are found near this atom.
which means the spin up/down does not have same number of orbitals. 
This is often because the Wannier functions are wrongly defined,
or badly localized. Please check the Wannier centers in the Wannier90 output file. 
"""
                    )
        self._spin_dict = {}
        self._atom_dict = {}
        for ispin, iatom in enumerate(self.ind_mag_atoms):
            self._spin_dict[iatom] = ispin
            self._atom_dict[ispin] = iatom

        self._prepare_orb_mmat()

    def _prepare_orb_mmat(self):
        self.mmats = {}
        self.orbital_names = {}
        self.norb_reduced = {}
        if self.backend_name == "SIESTA":
            syms = self.atoms.get_chemical_symbols()
            for iatom, orbs in self.labels.items():
                if (self.include_orbs is not None) and syms[iatom] in self.include_orbs:
                    mmat, reduced_orbs = map_orbs_matrix(
                        orbs,
                        spinor=not (self._is_collinear),
                        include_only=self.include_orbs[syms[iatom]],
                    )
                else:
                    mmat, reduced_orbs = map_orbs_matrix(
                        orbs, spinor=not (self._is_collinear), include_only=None
                    )

                self.mmats[iatom] = mmat
                self.orbital_names[iatom] = reduced_orbs
                # Note that for siesta, spin up and spin down has same orb name.
                # Therefor there is no nedd to /2
                self.norb_reduced[iatom] = len(reduced_orbs)
        else:
            self.orbital_names = self.labels
            for iatom, orbs in self.labels.items():
                # Note that for siesta, spin up and spin down has same orb name.
                # thus //2
                self.norb_reduced[iatom] = len(orbs) // 2

    def ispin(self, iatom):
        return self._spin_dict[iatom]

    def iatom(self, ispin):
        return self._atom_dict[ispin]

    def _prepare_distance(self):
        """
        prepare the distance between atoms.
        """
        self.distance_dict = {}
        self.short_Rlist = []
        self.R_ijatom_dict = defaultdict(lambda: [])
        ind_matoms = self.ind_mag_atoms
        for ispin, iatom in enumerate(ind_matoms):
            for jspin, jatom in enumerate(ind_matoms):
                for R in self.Rlist:
                    pos_i = self.atoms.get_positions()[iatom]
                    pos_jR = self.atoms.get_positions()[jatom] + np.dot(
                        R, self.atoms.get_cell()
                    )
                    vec = pos_jR - pos_i
                    distance = np.sqrt(np.sum(vec**2))
                    if self.Rcut is None or distance < self.Rcut:
                        self.distance_dict[(tuple(R), ispin, jspin)] = (vec, distance)
                        self.R_ijatom_dict[tuple(R)].append((iatom, jatom))
        self.short_Rlist = list(self.R_ijatom_dict.keys())

    def iorb(self, iatom):
        """
        given an index of atom, return all the orbital indices of it.
        :param iatom:  index of atom.
        """
        return np.array(self.orb_dict[iatom], dtype=int)

    def simplify_orbital_contributions(self, Jorbij, iatom, jatom):
        """
        sum up the contribution of all the orbitals with same (n,l,m)
        """
        if self.backend_name == "SIESTA":
            mmat_i = self.mmats[iatom]
            mmat_j = self.mmats[jatom]
            Jorbij = mmat_i.T @ Jorbij @ mmat_j
        return Jorbij

    def calculate_all(self):
        raise NotImplementedError(
            "calculate all is not implemented in a abstract class. "
        )

    def write_output(self):
        raise NotImplementedError(
            "write_output is not implemented in a abstract class. "
        )


class ExchangeNCL(Exchange):
    """
    Non-collinear exchange
    """

    def set_tbmodels(self, tbmodels):
        """
        tbmodels should be in spinor form.
        The basis should be orb1_up, orb2_up,...orbn_up, orb1_dn, orb2_dn....
        """
        self.tbmodel = tbmodels
        self.backend_name = self.tbmodel.name
        # TODO: check if tbmodels are really a tbmodel with SOC.
        self.G = TBGreen(
            self.tbmodel,
            self.kmesh,
            self.efermi,
            use_cache=self._use_cache,
            nproc=self.np,
        )
        self.norb = self.G.norb
        self.nbasis = self.G.nbasis
        # self.rho = np.zeros((self.nbasis, self.nbasis), dtype=complex)
        self.rho = self.G.get_density_matrix().real
        self.A_ijR_list = defaultdict(lambda: [])
        self.A_ijR = defaultdict(lambda: np.zeros((4, 4), dtype=complex))
        self.A_ijR_orb = dict()
        self.HR0 = self.G.H0
        self._is_collinear = False
        self.Pdict = {}
        if self.write_density_matrix:
            self.G.write_rho_R()

    def _prepare_NijR(self):
        self.N = {}
        for R in self.Rlist:
            self.N[R] = np.zeros((self.nbasis, self.nbasis), dtype=complex)

    def _prepare_Patom(self):
        for iatom in self.ind_mag_atoms:
            self.Pdict[iatom] = pauli_block_sigma_norm(self.get_H_atom(iatom))

    def get_H_atom(self, iatom):
        orbs = self.iorb(iatom)
        # return self.HR0[self.orb_slice[iatom], self.orb_slice[iatom]]
        return self.HR0[np.ix_(orbs, orbs)]

    def get_P_iatom(self, iatom):
        """Calculate the norm of the Hamiltonian vector.
        For each atom, the local hamiltonian of each orbital H(2*2 matrix)
        can be written as H0* I + H1* sigma1 + H2*sigma2 + H3 *sigma3
        where sigma is a Pauli matrix. return the norm of (H1, H2, H3) vector
        :param iatom: index of atom
        :returns: a matrix of norms P. P[i, j] is for orbital i and j.
        :rtype: complex matrix of shape norb_i * norb_i.
        """
        if self.Pdict == {}:
            self._prepare_Patom()
        return self.Pdict[iatom]

    def GR_atom(self, GR, iatom, jatom):
        """Given a green's function matrix, return the [iatom, jatom] component.

        :param GR:  Green's function matrix
        :param iatom:  index of atom i
        :param jatom:  index of atom j
        :returns:  G_ij
        :rtype:  complex matrix.

        """
        orbi = self.iorb(iatom)
        orbj = self.iorb(jatom)
        return GR[np.ix_(orbi, orbj)]
        # return GR[self.orb_slice[iatom], self.orb_slice[jatom]]

    def get_A_ijR(self, G, R, iatom, jatom):
        """calculate A from G for a energy slice (de).
        It take the
        .. math::
           A^{uv} = p T^u p T^v dE / pi

        where u, v are I, x, y, z (index 0, 1,2,3). p(i) = self.get_P_iatom(iatom)
        T^u(ijR)  (u=0,1,2,3) = pauli_block_all(G)

        :param G: Green's function for all R, i, j.
        :param iatom: i
        :param jatom: j
        :param de:  energy step. used for integeration
        :returns: a matrix of A_ij(u, v), where u, v =(0)0, x(1), y(2), z(3)
        :rtype:  4*4 matrix
        """
        GR = G[R]
        Gij = self.GR_atom(GR, iatom, jatom)
        Gij_Ixyz = pauli_block_all(Gij)

        # G(j, i, -R)
        Rm = tuple(-x for x in R)
        GRm = G[Rm]
        Gji = self.GR_atom(GRm, jatom, iatom)
        Gji_Ixyz = pauli_block_all(Gji)

        ni = self.norb_reduced[iatom]
        nj = self.norb_reduced[jatom]

        tmp = np.zeros((4, 4), dtype=complex)

        if self.orb_decomposition:
            torb = np.zeros((4, 4, ni, nj), dtype=complex)
            # a, b in (0,x,y,z)
            for a in range(4):
                piGij = self.get_P_iatom(iatom) @ Gij_Ixyz[a]
                for b in range(4):
                    pjGji = self.get_P_iatom(jatom) @ Gji_Ixyz[b]
                    torb[a, b] = self.simplify_orbital_contributions(
                        np.einsum("ij, ji -> ij", piGij, pjGji) / np.pi, iatom, jatom
                    )
                    tmp[a, b] = np.sum(torb[a, b])

        else:
            for a in range(4):
                pGp = self.get_P_iatom(iatom) @ Gij_Ixyz[a] @ self.get_P_iatom(jatom)
                for b in range(4):
                    AijRab = pGp @ Gji_Ixyz[b]
                    tmp[a, b] = np.trace(AijRab) / np.pi
            torb = None
        return tmp, torb

    def get_all_A(self, G):
        """
        Calculate all A matrix elements
        Loop over all magnetic atoms.
        :param G: Green's function.
        :param de: energy step.
        """
        A_ijR_list = {}
        Aorb_ijR_list = {}
        for iR, R in enumerate(self.R_ijatom_dict):
            for iatom, jatom in self.R_ijatom_dict[R]:
                A, A_orb = self.get_A_ijR(G, R, iatom, jatom)
                A_ijR_list[(R, iatom, jatom)] = A
                Aorb_ijR_list[(R, iatom, jatom)] = A_orb
        return A_ijR_list, Aorb_ijR_list

    def A_to_Jtensor_orb(self):
        """
        convert the orbital composition of A into J, DMI, Jani
        """
        self.Jiso_orb = {}
        self.Jani_orb = {}
        self.DMI_orb = {}

        if self.orb_decomposition:
            for key, val in self.A_ijR_orb.items():
                R, iatom, jatom = key
                Rm = tuple(-x for x in R)
                # valm = self.A_ijR_orb[(Rm, jatom, iatom)]
                ni = self.norb_reduced[iatom]
                nj = self.norb_reduced[jatom]

                is_nonself = not (R == (0, 0, 0) and iatom == jatom)
                ispin = self.ispin(iatom)
                jspin = self.ispin(jatom)
                keyspin = (R, ispin, jspin)

                # isotropic J
                Jiso = np.imag(val[0, 0] - val[1, 1] - val[2, 2] - val[3, 3])

                # off-diagonal anisotropic exchange
                Ja = np.zeros((3, 3, ni, nj), dtype=float)
                for i in range(3):
                    for j in range(3):
                        Ja[i, j] = np.imag(val[i + 1, j + 1] + val[j + 1, i + 1])
                # DMI

                Dtmp = np.zeros((3, ni, nj), dtype=float)
                for i in range(3):
                    Dtmp[i] = np.real(val[0, i + 1] - val[i + 1, 0])

                if is_nonself:
                    self.Jiso_orb[keyspin] = Jiso
                    self.Jani_orb[keyspin] = Ja
                    self.DMI_orb[keyspin] = Dtmp

    def A_to_Jtensor(self):
        """
        Calculate J tensors from A.
        If we assume the exchange can be written as a bilinear tensor form,
        J_{isotropic} = Tr Im (A^{00} - A^{xx} - A^{yy} - A^{zz})
        J_{anisotropic}_uv = Tr Im (2A)
        DMI =  Tr Re (A^{0z} - A^{z0} )
        """
        self.Jani = {}
        self.DMI = {}

        self.Jprime = {}
        self.B = {}
        self.exchange_Jdict = {}
        self.Jiso_orb = {}

        self.debug_dict = {"DMI2": {}}

        for key, val in self.A_ijR.items():
            # key:(R, iatom, jatom)
            R, iatom, jatom = key

            Rm = tuple(-x for x in R)
            valm = self.A_ijR[(Rm, jatom, iatom)]

            ispin = self.ispin(iatom)
            jspin = self.ispin(jatom)
            keyspin = (R, ispin, jspin)

            is_nonself = not (R == (0, 0, 0) and iatom == jatom)
            Jiso = 0.0
            Ja = np.zeros((3, 3), dtype=float)
            Dtmp = np.zeros(3, dtype=float)
            Dtmp2 = np.zeros(3, dtype=float)
            # Heisenberg like J.
            Jiso = np.imag(val[0, 0] - val[1, 1] - val[2, 2] - val[3, 3])

            if is_nonself:
                self.exchange_Jdict[keyspin] = Jiso

            # off-diagonal anisotropic exchange
            for i in range(3):
                for j in range(3):
                    Ja[i, j] = np.imag(val[i + 1, j + 1] + valm[i + 1, j + 1])
            if is_nonself:
                self.Jani[keyspin] = Ja

            # DMI
            for i in range(3):
                Dtmp[i] = np.real(val[0, i + 1] - val[i + 1, 0])

            # Dx = Jyz-Jzy
            # Dy = Jzx-Jxz
            # Dz = Jxy-Jyx
            Dtmp2[0] = np.imag(val[2, 3] - val[3, 2])
            Dtmp2[1] = np.imag(val[3, 1] - val[1, 3])
            Dtmp2[2] = np.imag(val[1, 2] - val[2, 1])
            if is_nonself:
                self.DMI[keyspin] = Dtmp
                self.debug_dict["DMI2"][keyspin] = Dtmp2

            # isotropic exchange into bilinear and biqudratic parts:
            # Jprime SiSj and B (SiSj)^2
            if is_nonself:
                Si = self.spinat[iatom]
                Sj = self.spinat[jatom]
                Jprime = np.imag(val[0, 0] - val[3, 3]) - 2 * np.sign(
                    np.dot(Si, Sj)
                ) * np.imag(val[3, 3])
                # Jprime = np.imag(val[0, 0] - 3*val[3, 3])
                B = np.imag(val[3, 3])
                self.B[keyspin] = Jprime, B

    # def get_N_e(self, GR, de):
    #    """
    #    calcualte density matrix for all R,i, j
    #    """
    #    self.N = defaultdict(lambda: 0.0)
    #    for R, G in GR.items():
    #        self.N[R] += -1.0 / np.pi * np.imag(G * de)

    # def get_rho_e(self, rhoR):
    #    """add component to density matrix from a green's function
    #    :param GR: Green's funciton in real space.
    #    """
    #    return -1.0 / np.pi * rhoR[0, 0, 0]

    # def get_total_charges(self):
    #    return np.sum(np.imag(np.diag(self.rho)))

    def get_rho_atom(self):
        """
        calculate charge and spin for each atom.
        """
        rho = {}
        self.charges = np.zeros(len(self.atoms), dtype=float)
        self.spinat = np.zeros((len(self.atoms), 3), dtype=float)
        for iatom in self.orb_dict:
            iorb = self.iorb(iatom)
            tmp = self.rho[np.ix_(iorb, iorb)]
            # *2 because there is a 1/2 in the paui_block_all function
            rho[iatom] = np.array([np.trace(x) * 2 for x in pauli_block_all(tmp)]).real
            self.charges[iatom] = rho[iatom][0]
            self.spinat[iatom, :] = rho[iatom][1:]
        self.rho_dict = rho
        return self.rho_dict

    def calculate_DMI_NJT(self):
        """
        calculate exchange and DMI with the
        D(i,j) =
        """
        Ddict_NJT = {}
        Jdict_NJT = {}
        for R in self.short_Rlist:
            N = self.N[tuple(-np.array(R))]  # density matrix
            t = self.tbmodel.get_hamR(R)  # hopping parameter
            for iatom in self.ind_mag_atoms:
                orbi = self.iorb(iatom)
                ni = len(orbi)
                for jatom in self.ind_mag_atoms:
                    orbj = self.iorb(jatom)
                    nj = len(orbj)
                    Nji = N[np.ix_(orbj, orbi)]
                    tij = t[np.ix_(orbi, orbj)]
                    D = np.zeros(3, dtype=float)
                    J = np.zeros(3, dtype=float)
                    for dim in range(3):
                        # S_i = pauli_mat(ni, dim +
                        #                1)  #*self.rho[np.ix_(orbi, orbi)]
                        # S_j = pauli_mat(nj, dim +
                        #                1)  #*self.rho[np.ix_(orbj, orbj)]
                        # TODO: Note that rho is complex, not the imaginary part
                        S_i = pauli_mat(ni, dim + 1) * self.rho[np.ix_(orbi, orbi)]
                        S_j = pauli_mat(nj, dim + 1) * self.rho[np.ix_(orbj, orbj)]

                        # [S, t]+  = Si tij + tij Sj, where
                        # Si and Sj are the spin operator
                        # Here we do not have L operator, so J-> S
                        Jt = np.matmul(S_i, tij) + np.matmul(tij, S_j)

                        Jtminus = np.matmul(S_i, tij) - np.matmul(tij, S_j)
                        # D = -1/2 Tr Nji [J, tij]
                        # Trace over spin and orb
                        D[dim] = -0.5 * np.imag(np.trace(np.matmul(Nji, Jt)))
                        J[dim] = -0.5 * np.imag(np.trace(np.matmul(Nji, Jtminus)))
                    ispin = self.ispin(iatom)
                    jspin = self.ispin(jatom)
                    Ddict_NJT[(R, ispin, jspin)] = D
                    Jdict_NJT[(R, ispin, jspin)] = J
        self.Jdict_NJT = Jdict_NJT
        self.Ddict_NJT = Ddict_NJT
        return Ddict_NJT

    def integrate(self, AijRs, AijRs_orb=None, method="simpson"):
        """
        AijRs: a list of AijR,
        wherer AijR: array of ((nR, n, n, 4,4), dtype=complex)
        """
        if method == "trapezoidal":
            integrate = trapezoidal_nonuniform
        elif method == "simpson":
            integrate = simpson_nonuniform

        # self.rho = integrate(self.contour.path, rhoRs)
        for iR, R in enumerate(self.R_ijatom_dict):
            for iatom, jatom in self.R_ijatom_dict[R]:
                f = AijRs[(R, iatom, jatom)]
                self.A_ijR[(R, iatom, jatom)] = integrate(self.contour.path, f)
                if self.orb_decomposition:
                    self.A_ijR_orb[(R, iatom, jatom)] = integrate(
                        self.contour.path, AijRs_orb[(R, iatom, jatom)]
                    )

    def get_AijR(self, e):
        GR = self.G.get_GR(self.short_Rlist, energy=e, get_rho=False)
        AijR, AijR_orb = self.get_all_A(GR)
        return AijR, AijR_orb

    def save_AijR(self, AijRs, fname):
        result = dict(path=self.contour.path, AijRs=AijRs)
        with open(fname, "wb") as myfile:
            pickle.dump(result, myfile)

    def validate(self):
        """
        Do some sanity check before proceding.
        """

    def calculate_all(self):
        """
        The top level.
        """
        print("Green's function Calculation started.")

        AijRs = {}

        AijRs_orb = {}

        self.validate()

        npole = len(self.contour.path)
        if self.np > 1:
            results = p_map(self.get_AijR, self.contour.path, num_cpus=self.np)
        else:
            results = map(self.get_AijR, tqdm(self.contour.path, total=npole))

        for i, result in enumerate(results):
            for iR, R in enumerate(self.R_ijatom_dict):
                for iatom, jatom in self.R_ijatom_dict[R]:
                    if (R, iatom, jatom) in AijRs:
                        AijRs[(R, iatom, jatom)].append(result[0][R, iatom, jatom])
                        if self.orb_decomposition:
                            AijRs_orb[(R, iatom, jatom)].append(
                                result[1][R, iatom, jatom]
                            )

                    else:
                        AijRs[(R, iatom, jatom)] = []
                        AijRs[(R, iatom, jatom)].append(result[0][R, iatom, jatom])
                        if self.orb_decomposition:
                            AijRs_orb[(R, iatom, jatom)] = []
                            AijRs_orb[(R, iatom, jatom)].append(
                                result[1][R, iatom, jatom]
                            )
        if self.np > 1:
            # executor.close()
            # executor.join()
            # executor.clear()
            pass

        # self.save_AijRs(AijRs)
        self.integrate(AijRs, AijRs_orb)
        self.get_rho_atom()
        self.A_to_Jtensor()
        self.A_to_Jtensor_orb()

    def _prepare_index_spin(self):
        # index_spin: index in spin hamiltonian of atom. starts from 1. -1 means not considered.
        ind_matoms = []
        self.index_spin = []
        ispin = 0
        symbols = self.atoms.get_chemical_symbols()
        tags = self.atoms.get_tags()
        for i, (sym, tag) in enumerate(zip(symbols, tags)):
            if sym in self.magnetic_elements or f"{sym}{tag}" in self.magnetic_elements:
                ind_matoms.append(i)
                self.index_spin.append(ispin)
                ispin += 1
            else:
                self.index_spin.append(-1)

    def write_output(self, path="TB2J_results"):
        self._prepare_index_spin()
        output = SpinIO(
            atoms=self.atoms,
            charges=self.charges,
            spinat=self.spinat,
            index_spin=self.index_spin,
            colinear=False,
            orbital_names=self.orbital_names,
            distance_dict=self.distance_dict,
            exchange_Jdict=self.exchange_Jdict,
            Jiso_orb=self.Jiso_orb,
            dmi_ddict=self.DMI,
            Jani_dict=self.Jani,
            DMI_orb=self.DMI_orb,
            Jani_orb=self.Jani_orb,
            NJT_Jdict=self.Jdict_NJT,
            NJT_ddict=self.Ddict_NJT,
            biquadratic_Jdict=self.B,
            debug_dict=self.debug_dict,
            description=self.description,
        )
        output.write_all(path=path)
        # with open("TB2J_results/J_orb.pickle", 'wb') as myfile:
        #    pickle.dump({'Jiso_orb': self.Jiso_orb,
        #                 'DMI_orb': self.DMI_orb, 'Jani_orb': self.Jani_orb}, myfile)

    def finalize(self):
        self.G.clean_cache()

    def run(self, path="TB2J_results"):
        self.calculate_all()
        self.write_output(path=path)
        self.finalize()


class ExchangeCL(ExchangeNCL):
    def set_tbmodels(self, tbmodel):
        """
        only difference is a colinear tag.
        """
        super().set_tbmodels(tbmodel)
        self._is_collinear = True

    def write_output(self, path="TB2J_results"):
        self._prepare_index_spin()
        output = SpinIO(
            atoms=self.atoms,
            charges=self.charges,
            spinat=self.spinat,
            index_spin=self.index_spin,
            orbital_names=self.orbital_names,
            colinear=True,
            distance_dict=self.distance_dict,
            exchange_Jdict=self.exchange_Jdict,
            dmi_ddict=None,
            NJT_Jdict=None,
            NJT_ddict=None,
            biquadratic_Jdict=self.B,
            description=self.description,
        )
        output.write_all(path=path)



"""
Hamiltonian terms
"""
import numpy as np
from collections import defaultdict
import scipy.sparse as ssp


class HamTerm(object):
    """
    Base class for Hamiltonian terms.
    """

    def __init__(self, ms=None):
        self.E = 0.0
        self.ms = ms
        if ms is not None:
            self.nmatoms = len(ms)
        self._hessian = None
        self._hessian_ijR = None

    def func(self):
        raise NotImplementedError()

    def eff_field(self, S):
        r"""
        Hi = -1/ms_i * (\partial H/\partial Si)
        """
        return -self.jacobian(S)

    def calc_hessian(self):
        raise NotImplementedError()

    def hessian(self):
        r"""
        \partial^2 H/ (\paritial Si \partial Sj)
        """
        if self._hessian is None:
            self.calc_hessian()
        return self._hessian

    def calc_hessian_ijR(self):
        raise NotImplementedError()

    def hessian_ijR(self):
        if self._hessian_ijR is None:
            self.calc_hessian_ijR()
        return self._hessian_ijR

    def energy(self, S):
        raise NotImplementedError()


class SingleBodyTerm(HamTerm):
    def __init__(self, ms=None):
        super(SingleBodyTerm, self).__init__(ms=ms)

    def func(self):
        return np.sum((self.func_i for i in range(self.nmatoms)))

    def func_i(self):
        pass

    def jacobian(self, S):
        return np.array([self.jacobian_i(S, i) for i in range(self.nmatoms)])

    def jacobian_i(self, S, i):
        pass

    def calc_hessian(self):
        self._hessian = 0

    def calc_hessian_ijR(self):
        self._hessian_ijR = {}

    def is_twobody_term(self):
        return False


class TwoBodyTerm(HamTerm):
    def __init__(self, ms=None):
        super(TwoBodyTerm, self).__init__(ms=ms)

    def func(self, S):
        E = 0.0
        for i, j in self.pair_list:
            if i != j:
                E += self.func_ij(S, i, j)
        return E

    def func_ij(self, i, j):
        raise NotImplementedError()

    def jacobian(self, S):
        r"""
        \partial H/\partial Si
        """
        raise NotImplementedError()

    def jacobian_i(self, S, i):
        raise NotImplementedError()

    def is_twobody_term(self):
        return True


class ZeemanTerm(SingleBodyTerm):
    """
    $H_{zeeman}=-\\sum_i g_i \\mu_B \\vec{H}_i \\cdot \\vec{S}_i $
    """

    def __init__(self, H, ms=None):
        super(ZeemanTerm, self).__init__(ms)
        self.H = H

    def eff_field(self, S, Heff):
        r"""
        Hi = -1/ms_i * (\partial H/\partial Si)
        It is here because it is simpler than the form of jacobian. Therefore faster.
        """
        Heff[:, :] += self.H * self.ms[:, None]


class UniaxialMCATerm(TwoBodyTerm):
    """ """

    def __init__(self, k1, k1dir, ms):
        super(UniaxialMCATerm, self).__init__(ms=ms)
        self.Ku = np.array(k1)  # Ku list of nspins.
        # normalize direction.
        direction = np.array(k1dir)
        self.e = direction / (np.linalg.norm(direction, axis=1)[:, None])
        assert self.e.shape == (self.nmatoms, 3)
        assert self.Ku.shape == ms.shape

    def func_i(self, S, i):
        return -self.Ku[i] * np.dot(S[i], self.e[i]) ** 2

    def eff_field(self, S, Heff):
        Heff[:, :] -= (
            self.hessian().dot(S.reshape(3 * self.nmatoms)).reshape(self.nmatoms, 3)
        )

    def calc_hessian(self):
        self._hessian = ssp.lil_matrix(
            (self.nmatoms * 3, self.nmatoms * 3), dtype=float
        )
        for i in range(self.nmatoms):
            self._hessian[i * 3 : i * 3 + 3, i * 3 : i * 3 + 3] = (
                -2.0 * self.Ku[i]
            ) * np.outer(self.e[i], self.e[i])
        self._hessian = ssp.csr_matrix(self._hessian)
        return self._hessian

    def calc_hessian_ijR(self):
        self._hessian_ijR = {}
        for i in range(self.nmatoms):
            self._hessian_ijR[(i, i, (0, 0, 0))] = (-2.0 * self.Ku[i]) * np.outer(
                self.e[i], self.e[i]
            )
        return self._hessian_ijR


class HomoUniaxialMCATerm(SingleBodyTerm):
    """
    Homogenous Uniaxial Magnetocrystaline Anistropy
    """

    def __init__(self, Ku, direction, ms=None):
        super(HomoUniaxialMCATerm, self).__init__(ms=ms)
        self.Ku = Ku
        self.e = np.array(direction) / np.linalg.norm(direction)

    def func_i(self, S, i):
        return -self.Ku * np.dot(S[i], self.e) ** 2

    def jacobian_i(self, S, i):
        return -2.0 * self.Ku * np.dot(S[i], self.e) * self.e

    def eff_field(self, S, Heff):
        Heff += 2.0 * self.Ku * np.outer(np.einsum("ij,j,i->i", S, self.e, 1), self.e)


class ExchangeTerm(TwoBodyTerm):
    """
    exchane interaction in Heissenberg model
    """

    def __init__(self, Jdict, ms=None, sparse_matrix_form=True, pbc=[1, 1, 1]):
        """
        J is given as a dict of {(i, j, R): val},
         where R is a tuple, val is a scalar.
        """
        super(ExchangeTerm, self).__init__(ms=ms)
        self.Jdict = Jdict
        Jmat = defaultdict(float)
        for key, val in self.Jdict.items():
            i, j, R = key
            if np.all(R * (1 - np.array(pbc)) == 0):
                Jmat[(i, j)] += val
        self.ilist, self.jlist = np.array(tuple(Jmat.keys()), dtype="int").T
        self.vallist = np.array(tuple(Jmat.values()))
        self.jac = np.zeros((self.nmatoms, 3))
        self.nij = self.vallist.shape[0]
        self.Heff = np.zeros((self.nmatoms, 3))

    def jacobian(self, S):
        self.jac = -2.0 * self.hessian().dot(S.reshape(self.nmatoms * 3)).reshape(
            self.nmatoms, 3
        )
        return self.jac

    def calc_hessian(self):
        self._hessian = ssp.lil_matrix(
            (self.nmatoms * 3, self.nmatoms * 3), dtype=float
        )
        for i, j, val in zip(self.ilist, self.jlist, self.vallist):
            self._hessian[i * 3 : i * 3 + 3, j * 3 : j * 3 + 3] += np.eye(3) * val

        self._hessian = ssp.csr_matrix(self._hessian)
        return self._hessian

    def calc_hessian_ijR(self):
        self._hessian_ijR = {}
        for key, val in self.Jdict.items():
            i, j, R = key
            self._hessian_ijR[(i, j, R)] = np.eye(3) * val
        return self._hessian_ijR


class DMITerm(TwoBodyTerm):
    """
    Dzyaloshinskii-Moriya interaction.
    $H_{DM} = -\\sum {i<j} \\vec{D}_{ij} \\cdot \\S_i        imes \\S_j$
    """

    def __init__(self, ddict, ms):
        """
        J is given as a dict of {(i, j, R): val},
         where R is a tuple, val is a scalar.
        """
        super(DMITerm, self).__init__(ms=ms)
        self.ddict = ddict
        Dmat = defaultdict(float)
        for key, val in self.ddict.items():
            i, j, R = key
            Dmat[(i, j)] += np.array(val)

        self.ilist, self.jlist = np.array(tuple(Dmat.keys()), dtype="int").T
        self.vallist = np.array(tuple(Dmat.values()))
        self.jac = np.zeros((self.nmatoms, 3))
        self.nij = self.vallist.shape[0]
        self.Heff = np.zeros((self.nmatoms, 3))

    def jacobian(self, S):
        return self.hessian().dot(S.reshape(3 * self.nmatoms)).reshape(self.nmatoms, 3)

    def calc_hessian(self):
        # self._hessian = np.zeros(self.nmatoms * 3, self.nmatoms * 3)
        self._hessian = ssp.lil_matrix(
            (self.nmatoms * 3, self.nmatoms * 3), dtype=float
        )
        for i, j, val in zip(self.ilist, self.jlist, self.vallist):
            self._hessian[i * 3 : i * 3 + 3, j * 3 : j * 3 + 3] += np.array(
                [[0, val[2], -val[1]], [-val[2], 0, val[0]], [val[1], -val[0], 0]]
            )
        self._hessian = ssp.csr_matrix(self._hessian)
        return self._hessian

    def calc_hessian_ijR(self):
        self._hessian_ijR = {}
        for key, val in self.ddict.items():
            i, j, R = key
            self._hessian_ijR[(i, j, R)] = np.array(
                [[0, val[2], -val[1]], [-val[2], 0, val[0]], [val[1], -val[0], 0]]
            )
        return self._hessian_ijR


class BilinearTerm(TwoBodyTerm):
    """
    Bilinear term
    """

    def __init__(self, bidict, ms):
        """
        J is given as a dict of {(i, j, R): val},
         where R is a tuple, val is a scalar.
        """
        super(BilinearTerm, self).__init__(ms=ms)
        self.bidict = bidict
        bimat = defaultdict(float)
        for key, val in self.bidict.items():
            i, j, R = key
            bimat[(i, j)] += np.array(val)

        self.ilist, self.jlist = np.array(tuple(bimat.keys()), dtype="int").T
        self.vallist = np.array(tuple(bimat.values()))
        self.jac = np.zeros((self.nmatoms, 3))
        self.nij = self.vallist.shape[0]
        self.Heff = np.zeros((self.nmatoms, 3))

    def jacobian(self, S):
        return self.hessian().dot(S.reshape(3 * self.nmatoms)).reshape(self.nmatoms, 3)

    def calc_hessian(self):
        # self._hessian = np.zeros(self.nmatoms * 3, self.nmatoms * 3)
        self._hessian = ssp.lil_matrix(
            (self.nmatoms * 3, self.nmatoms * 3), dtype=float
        )
        for i, j, val in zip(self.ilist, self.jlist, self.vallist):
            self._hessian[i * 3 : i * 3 + 3, j * 3 : j * 3 + 3] += np.array(val)
        self._hessian = ssp.csr_matrix(self._hessian)
        return self._hessian

    def calc_hessian_ijR(self):
        self._hessian_ijR = {}
        for key, val in self.bidict.items():
            i, j, R = key
            self._hessian_ijR[(i, j, R)] = np.array(val)
        return self._hessian_ijR


class DipDip(TwoBodyTerm):
    """
    Dipolar interaction.
    TODO Note that Model.positions is reduced coordinates.
    """

    def __init__(self):
        pass



import numpy as np
from collections import defaultdict
from ase.dft.kpoints import monkhorst_pack
from shutil import rmtree
import os
import tempfile
from pathos.multiprocessing import ProcessPool
import sys
import pickle

MAX_EXP_ARGUMENT = np.log(sys.float_info.max)


def eigen_to_G(evals, evecs, efermi, energy):
    """calculate green's function from eigenvalue/eigenvector for energy(e-ef): G(e-ef).
    :param evals:  eigen values
    :param evecs:  eigen vectors
    :param efermi: fermi energy
    :param energy: energy
    :returns: Green's function G,
    :rtype:  Matrix with same shape of the Hamiltonian (and eigenvector)
    """
    return (
        np.einsum("ij, j-> ij", evecs, 1.0 / (-evals + (energy + efermi)))
        @ evecs.conj().T
    )


def fermi(e, mu, width=0.01):
    """
    the fermi function.
     .. math::
        f=\\frac{1}{\exp((e-\mu)/width)+1}

    :param e,mu,width: e,\mu,width
    """

    x = (e - mu) / width
    return np.where(x < MAX_EXP_ARGUMENT, 1 / (1.0 + np.exp(x)), 0.0)


def find_energy_ingap(evals, rbound, gap=4.0):
    """
    find a energy inside a gap below rbound (right bound),
    return the energy gap top - 0.5.
    """
    m0 = np.sort(evals.flatten())
    m = m0[m0 < rbound]
    ind = np.where(np.diff(m) > gap)[0]
    if len(ind) == 0:
        return m0[0] - 0.5
    else:
        return m[ind[-1] + 1] - 0.5


class TBGreen:
    def __init__(
        self,
        tbmodel,
        kmesh,  # [ikpt, 3]
        efermi,  # efermi
        k_sym=False,
        use_cache=False,
        cache_path=None,
        nproc=1,
    ):
        """
        :param tbmodel: A tight binding model
        :param kmesh: size of monkhorst pack. e.g [6,6,6]
        :param efermi: fermi energy.
        """
        self.tbmodel = tbmodel
        self.is_orthogonal = tbmodel.is_orthogonal
        self.R2kfactor = tbmodel.R2kfactor
        self.k2Rfactor = -tbmodel.R2kfactor
        self.efermi = efermi
        self._use_cache = use_cache
        self.cache_path = cache_path
        if use_cache:
            self._prepare_cache()
        if kmesh is not None:
            self.kpts = monkhorst_pack(size=kmesh)
        else:
            self.kpts = tbmodel.get_kpts()
        self.nkpts = len(self.kpts)
        self.kweights = [1.0 / self.nkpts] * self.nkpts
        self.norb = tbmodel.norb
        self.nbasis = tbmodel.nbasis
        self.k_sym = k_sym
        self.nproc = nproc
        self._prepare_eigen()

    def _reduce_eigens(self, evals, evecs, emin, emax):
        ts = np.logical_and(evals >= emin, evals < emax)
        ts = np.any(ts, axis=0)
        ts = np.where(ts)[0]
        if len(ts) == 0:
            raise ValueError(
                f"Cannot find any band in the energy range specified by emin {emin} and emax {emax}, which are relative to the Fermi energy. Please check that the Fermi energy, the emin and emax are correct. If you're using Wannier90 output, check the Wannier functions give the right band structure."
            )
        istart, iend = ts[0], ts[-1] + 1
        return evals[:, istart:iend], evecs[:, :, istart:iend]

    def find_energy_ingap(self, rbound, gap=4.0):
        return find_energy_ingap(self.evals, rbound, gap)

    def _prepare_cache(self):
        if self.cache_path is None:
            if "TMPDIR" in os.environ:
                rpath = os.environ["TMPDIR"]
            else:
                rpath = "/dev/shm/TB2J_cache"
        else:
            rpath = self.cache_path
        if not os.path.exists(rpath):
            os.makedirs(rpath)
        self.cache_path = tempfile.mkdtemp(prefix="TB2J", dir=rpath)
        print(f"Writting wavefunctions and Hamiltonian in cache {self.cache_path}")

    def clean_cache(self):
        if (self.cache_path is not None) and os.path.exists(self.cache_path):
            rmtree(self.cache_path)

    def _prepare_eigen(self, solve=True, saveH=False):
        """
        calculate eigen values and vectors for all kpts and save.
        Note that the convention 2 is used here, where the
        phase factor is e^(ik.R), not e^(ik.(R+rj-ri))
        """
        nkpts = len(self.kpts)
        self.evals = np.zeros((nkpts, self.nbasis), dtype=float)
        self.nkpts = nkpts
        self.H0 = np.zeros((self.nbasis, self.nbasis), dtype=complex)
        self.evecs = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
        self.H = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
        if not self.is_orthogonal:
            self.S = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
        else:
            self.S = None
        if self.nproc == 1:
            results = map(self.tbmodel.HSE_k, self.kpts)
        else:
            executor = ProcessPool(nodes=self.nproc)
            results = executor.map(self.tbmodel.HSE_k, self.kpts, [2] * len(self.kpts))
            executor.close()
            executor.join()
            executor.clear()

        for ik, result in enumerate(results):
            if self.is_orthogonal:
                self.H[ik], _, self.evals[ik], self.evecs[ik] = result
            else:
                self.H[ik], self.S[ik], self.evals[ik], self.evecs[ik] = result
            self.H0 += self.H[ik] / self.nkpts

        if not saveH:
            self.H = None

        self.evals, self.evecs = self._reduce_eigens(
            self.evals, self.evecs, emin=self.efermi - 10.0, emax=self.efermi + 10.1
        )
        if self._use_cache:
            evecs = self.evecs
            self.evecs_shape = self.evecs.shape
            self.evecs = np.memmap(
                os.path.join(self.cache_path, "evecs.dat"),
                mode="w+",
                shape=self.evecs.shape,
                dtype=complex,
            )
            self.evecs[:, :, :] = evecs[:, :, :]
            del self.evecs

            if self.is_orthogonal:
                self.S = None
            else:
                S = self.S
                self.S = np.memmap(
                    os.path.join(self.cache_path, "S.dat"),
                    mode="w+",
                    shape=(nkpts, self.nbasis, self.nbasis),
                    dtype=complex,
                )
                self.S[:] = S[:]
            if not self.is_orthogonal:
                del self.S

    def get_evecs(self, ik):
        if self._use_cache:
            return np.memmap(
                os.path.join(self.cache_path, "evecs.dat"),
                mode="r",
                shape=self.evecs_shape,
                dtype=complex,
            )[ik]
        else:
            return self.evecs[ik]

    def get_evalue(self, ik):
        return self.evals[ik]

    def get_Hk(self, ik):
        if self._use_cache:
            return np.memmap(
                os.path.join(self.cache_path, "H.dat"),
                mode="r",
                shape=(self.nkpts, self.nbasis, self.nbasis),
                dtype=complex,
            )[ik]
        else:
            return self.H[ik]

    def get_Sk(self, ik):
        if self.is_orthogonal:
            return None
        elif self._use_cache:
            return np.memmap(
                os.path.join(self.cache_path, "S.dat"),
                mode="r",
                shape=(self.nkpts, self.nbasis, self.nbasis),
                dtype=complex,
            )[ik]
        else:
            return self.S[ik]

    def get_density_matrix(self):
        rho = np.zeros((self.nbasis, self.nbasis), dtype=complex)
        for ik, _ in enumerate(self.kpts):
            rho += (
                (self.get_evecs(ik) * fermi(self.evals[ik], self.efermi))
                @ self.get_evecs(ik).T.conj()
                * self.kweights[ik]
            )
        return rho

    def get_rho_R(self, Rlist):
        nR = len(Rlist)
        rho_R = np.zeros((nR, self.nbasis, self.nbasis), dtype=complex)
        for ik, kpt in enumerate(self.kpts):
            evec = self.get_evecs(ik)
            rhok = np.einsum(
                "ib,b, bj-> ij", evec, fermi(self.evals[ik], self.efermi), evec.conj().T
            )
            for iR, R in enumerate(Rlist):
                rho_R[iR] += rhok * np.exp(self.k2Rfactor * kpt @ R) * self.kweights[ik]
        return rho_R

    def write_rho_R(self, Rlist, fname="rhoR.pickle"):
        rho_R = self.get_rho_R(Rlist)
        with open(fname, "wb") as myfile:
            pickle.dump({"Rlist": Rlist, "rhoR": rho_R}, myfile)

    def get_density(self):
        return np.real(np.diag(self.get_density_matrix()))

    def get_Gk(self, ik, energy):
        """Green's function G(k) for one energy
        G(\epsilon)= (\epsilon I- H)^{-1}
        :param ik: indices for kpoint
        :returns: Gk
        :rtype:  a matrix of indices (nbasis, nbasis)
        """
        Gk = eigen_to_G(
            evals=self.get_evalue(ik),
            evecs=self.get_evecs(ik),
            efermi=self.efermi,
            energy=energy,
        )

        # A slower version. For test.
        # Gk = np.linalg.inv((energy+self.efermi)*self.S[ik,:,:] - self.H[ik,:,:])
        return Gk

    def get_GR(self, Rpts, energy, get_rho=False):
        """calculate real space Green's function for one energy, all R points.
        G(R, epsilon) = G(k, epsilon) exp(-2\pi i R.dot. k)
        :param Rpts: R points
        :param energy:
        :returns:  real space green's function for one energy for a list of R.
        :rtype:  dictionary, the keys are tuple of R, values are matrices of nbasis*nbasis
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0j)
        rhoR = defaultdict(lambda: 0.0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            if get_rho:
                if self.is_orthogonal:
                    rhok = Gk
                else:
                    rhok = self.get_Sk(ik) @ Gk
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                tmp = Gk * (phase * self.kweights[ik])
                GR[R] += tmp
                # change this if need full rho
                if get_rho and R == (0, 0, 0):
                    rhoR[R] += rhok * (phase * self.kweights[ik])
        if get_rho:
            return GR, rhoR
        else:
            return GR

    def get_GR_and_dGRdx1(self, Rpts, energy, dHdx):
        """
        calculate G(R) and dG(R)/dx.
        dG(R)/dx = \sum_k G(k) (dH(R)/dx) G(k).
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0 + 0.0j)
        dGRdx = defaultdict(lambda: 0.0 + 0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            Gkw = Gk * self.kweights[ik]
            # Gmk = self.get_Gk(self.i_minus_k(kpt), energy)
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                GR[R] += Gkw * (phase * self.kweights[ik])

                dHRdx = dHdx.get_hamR(R)
                dGRdx[R] += Gkw @ dHRdx @ Gk
                # dGRdx[R] += Gk.dot(dHRdx).dot(Gkp)
        return GR, dGRdx

    def get_GR_and_dGRdx(self, Rpts, energy, dHdx):
        """
        calculate G(R) and dG(R)/dx.
        dG(k)/dx =  G(k) (dH(k)/dx) G(k).
        dG(R)/dx = \sum_k dG(k)/dx * e^{-ikR}
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0 + 0.0j)
        dGRdx = defaultdict(lambda: 0.0 + 0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            # Gmk = self.get_Gk(self.i_minus_k(kpt), energy)
            Gkp = Gk * self.kweights[ik]
            dHk = dHdx.gen_ham(tuple(kpt))
            dG = Gk @ dHk @ Gkp
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                GR[R] += Gkp * (phase * self.kweights[ik])
                dGRdx[R] += dG * (phase * self.kweights[ik])
        return GR, dGRdx

    def get_GR_and_dGRdx_and_dGRdx2(self, Rpts, energy, dHdx, dHdx2):
        """
        calculate G(R) and dG(R)/dx.
        dG(k)/dx =  G(k) (dH(k)/dx) G(k).
        dG(R)/dx = \sum_k dG(k)/dx * e^{-ikR}
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0 + 0.0j)
        dGRdx = defaultdict(lambda: 0.0 + 0j)
        dGRdx2 = defaultdict(lambda: 0.0 + 0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            # Gmk = self.get_Gk(self.i_minus_k(kpt), energy)
            Gkp = Gk * self.kweights[ik]
            dHk = dHdx.gen_ham(tuple(kpt))
            dHk2 = dHdx2.gen_ham(tuple(kpt))
            dG = Gk @ dHk @ Gkp
            dG2 = Gk @ dHk2 @ Gkp
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                GR[R] += Gkp * (phase * self.kweights[ik])
                dGRdx[R] += dG * (phase * self.kweights[ik])
                dGRdx2[R] += dG2 * (phase * self.kweights[ik])
        return GR, dGRdx, dGRdx2



import os
import copy
import numpy as np
from scipy.spatial.transform import Rotation
from TB2J.io_exchange import SpinIO
from TB2J.tensor_rotate import remove_components
from TB2J.Jtensor import DMI_to_Jtensor, Jtensor_to_DMI
from TB2J.tensor_rotate import Rzx, Rzy, Rzz, Ryz, Rxz


def test_rotation_matrix():
    x = [1, 0, 0]
    y = [0, 1, 0]
    z = [0, 0, 1]
    print(Rxz.apply(x))
    print(Ryz.apply(y))
    print(Rzx.apply(z))
    print(Rzy.apply(z))


def recover_DMI_from_rotated_structure(Ddict, rotation):
    """
    Recover the DMI vector from the rotated structure.
    D: the dictionary of DMI vector in the rotated structure.
    rotation: the rotation operator from the original structure to the rotated structure.
    """
    for key, val in Ddict.items():
        Ddict[key] = rotation.apply(val, inverse=True)
    return Ddict


def recover_Jani_fom_rotated_structure(Janidict, rotation):
    """
    Recover the Jani tensor from the rotated structure.
    Janidict: the dictionary of Jani tensor in the rotated structure.
    rotation: the  from the original structure to the rotated structure.
    """
    R = rotation.as_matrix()
    RT = R.T
    for key, Jani in Janidict.items():
        # Note: E=Si J Sj , Si'= Si RT, Sj' = R Sj,
        #       Si' J' Sj' = Si RT R J RT R Sj => J' = R J RT
        #       But here we are doing the opposite rotation back to
        #       the original axis, so we replace R with RT.
        Janidict[key] = RT @ Jani @ R
    return Janidict


def recover_spinat_from_rotated_structure(spinat, rotation):
    """
    Recover the spinat from the rotated structure.
    spinat: the spinat in the rotated structure.
    rotation: the rotation operator from the original structure to the rotated structure.
    """
    for i, spin in enumerate(spinat):
        spinat[i] = rotation.apply(spin, inverse=True)
    return spinat


# test_rotation_matrix()

# R_xyz = [Rxz.as_matrix(), Ryz.as_matrix(), np.eye(3, dtype=float)]


def rot_merge_DMI(Dx, Dy, Dz):
    Dx_z = Rzx.apply(Dx)
    Dy_z = Rzy.apply(Dy)
    D = (
        np.array([0.0, 0.5, 0.5]) * Dx_z
        + np.array([0.5, 0.0, 0.5]) * Dy_z
        + np.array([0.5, 0.5, 0.0]) * Dz
    )
    return D


def rot_merge_DMI2(Dx, Dy, Dz):
    Dx_z = Rzx.apply(Dx)
    Dy_z = Rzy.apply(Dy)
    D = (
        np.array([1, 0, 0]) * Dx_z
        + np.array([0, 1, 0]) * Dy_z
        + np.array([0, 0, 1]) * Dz
    )
    return D


def merge_DMI(Dx, Dy, Dz):
    D = (
        np.array([0.0, 0.5, 0.5]) * Dx
        + np.array([0.5, 0.0, 0.5]) * Dy
        + np.array([0.5, 0.5, 0.0]) * Dz
    )
    return D


def merge_DMI2(Dx, Dy, Dz):
    D = np.array([1, 0, 0]) * Dx + np.array([0, 1, 0]) * Dy + np.array([0, 0, 1]) * Dz
    return D


def swap_direction(m, idirections):
    """
    swap two directions of a tensor m.
    idirections: the index of the two directions.
    """
    idirections = list(idirections)
    inv = [idirections[1], idirections[0]]
    n = np.copy(m)
    n[:, idirections] = n[:, inv]
    n[idirections, :] = n[inv, :]
    return n


def test_swap():
    m = np.reshape(np.arange(9), (3, 3))
    print(m)
    print(swap_direction(m, [0, 1]))


def merge_Jani(Janix, Janiy, Janiz):
    # This is wrong.
    # Jani = (
    #    np.array([[0, 0, 0], [0, 1, 1], [0, 1, 1]]) * Janix
    #    + np.array([[1, 0, 1], [0, 0, 0], [1, 0, 1]]) * Janiy
    #    + np.array([[1, 1, 0], [1, 1, 0], [0, 0, 0]]) * Janiz
    # ) / 2.0
    wx = np.array([[0, 0, 0], [0, 1, 1], [0, 1, 1]])
    wy = np.array([[1, 0, 1], [0, 0, 0], [1, 0, 1]])
    wz = np.array([[1, 1, 0], [1, 1, 0], [0, 0, 0]])
    Jani = (wx * Janix + wy * Janiy + wz * Janiz) / (wx + wy + wz)
    return Jani


def read_pickle(path):
    p1 = os.path.join(path, "TB2J_results", "TB2J.pickle")
    p2 = os.path.join(path, "TB2J.pickle")
    if os.path.exists(p1) and os.path.exists(p2):
        print(f" WARNING!: Both file {p1} and {p2} exist. Use default {p1}.")
    if os.path.exists(p1):
        ret = SpinIO.load_pickle(os.path.join(path, "TB2J_results"))
    elif os.path.exists(p2):
        ret = SpinIO.load_pickle(path)
    else:
        raise FileNotFoundError(f"Cannot find either file {p1} or {p2}")
    return ret


class Merger2:
    def __init__(self, paths, method):
        self.method = method
        if method.lower() == "spin":
            self.load_with_rotated_spin(paths)
        elif method.lower() == "structure":
            self.load_with_rotated_structure(paths)
        else:
            raise ValueError("method should be either 'spin' or 'structure'")

    def load_with_rotated_structure(self, paths):
        """
        Merge TB2J results from multiple calculations.
        :param paths: a list of paths to the TB2J results.
        :param method: 'structure' or 'spin'
        """
        self.paths = paths
        if len(self.paths) != 3:
            raise ValueError(
                "The number of paths should be 3, with structure rotated from z to x, y, z"
            )
        for i, path in enumerate(self.paths):
            read_pickle(path)
        self.indata = [read_pickle(path) for path in paths]

        self.dat = copy.deepcopy(self.indata[-1])
        # self.dat.description += (
        #    "Merged from TB2J results in paths: \n  " + "\n  ".join(paths) + "\n"
        # )
        Rotations = [Rzx, Rzy, Rzz]
        for dat, rotation in zip(self.indata, Rotations):
            dat.spinat = recover_spinat_from_rotated_structure(dat.spinat, rotation)
            dat.dmi_ddict = recover_DMI_from_rotated_structure(dat.dmi_ddict, rotation)
            dat.Jani_dict = recover_Jani_fom_rotated_structure(dat.Jani_dict, rotation)

    def load_with_rotated_spin(self, paths):
        """
        Merge TB2J results from multiple calculations.
        :param paths: a list of paths to the TB2J results.
        :param method: 'structure' or 'spin'
        """
        self.paths = paths
        self.indata = [read_pickle(path) for path in paths]
        self.dat = copy.deepcopy(self.indata[-1])
        # self.dat.description += (
        #    "Merged from TB2J results in paths: \n  " + "\n  ".join(paths) + "\n"
        # )

    def merge_Jani(self):
        """
        Merge the anisotropic exchange tensor.
        """
        Jani_dict = {}
        for key, Jani in self.dat.Jani_dict.items():
            R, i, j = key
            weights = np.zeros((3, 3), dtype=float)
            Jani_sum = np.zeros((3, 3), dtype=float)
            for dat in self.indata:
                Si = dat.get_spin_ispin(i)
                Sj = dat.get_spin_ispin(j)
                # print(f"{Si=}, {Sj=}")
                Jani = dat.get_Jani(i, j, R, default=np.zeros((3, 3), dtype=float))
                Jani_removed, w = remove_components(
                    Jani,
                    Si,
                    Sj,
                    remove_indices=[[0, 2], [1, 2], [2, 2], [2, 1], [2, 0]],
                )
                w = Jani_removed / Jani
                Jani_sum += Jani * w  # Jani_removed
                # print(f"{Jani* w=}")
                weights += w
            # print(f"{weights=}")
            if np.any(weights == 0):
                raise RuntimeError(
                    "The data set to be merged does not give a complete anisotropic J tensor, please add more data"
                )
            Jani_dict[key] = Jani_sum / weights
        self.dat.Jani_dict = Jani_dict

    def merge_DMI(self):
        """
        merge the DMI vector
        """
        DMI = {}
        for key, D in self.dat.dmi_ddict.items():
            R, i, j = key
            weights = np.zeros((3, 3), dtype=float)
            Dtensor_sum = np.zeros((3, 3), dtype=float)
            for dat in self.indata:
                Si = dat.get_spin_ispin(i)
                Sj = dat.get_spin_ispin(j)
                D = dat.get_DMI(i, j, R, default=np.zeros((3,), dtype=float))
                Dtensor = DMI_to_Jtensor(D)
                Dtensor_removed, w = remove_components(
                    Dtensor, Si, Sj, remove_indices=[[0, 1], [1, 0]]
                )
                Dtensor_sum += Dtensor * w  # Dtensor_removed
                weights += w
            if np.any(weights == 0):
                raise RuntimeError(
                    "The data set to be merged does not give a complete DMI vector, please add more data"
                )
            DMI[key] = Jtensor_to_DMI(Dtensor_sum / weights)
        self.dat.dmi_ddict = DMI

    def merge_Jiso(self):
        """
        merge the isotropic exchange
        """
        Jiso = {}
        for key, J in self.dat.exchange_Jdict.items():
            R, i, j = key
            weights = 0.0
            Jiso_sum = 0.0
            for dat in self.indata:
                Si = dat.get_spin_ispin(i)
                Sj = dat.get_spin_ispin(j)
                J = dat.get_Jiso(i, j, R, default=0.0)
                Jiso_sum += J  # *np.eye(3, dtype=float)
                weights += 1.0
            if np.any(weights == 0):
                raise RuntimeError(
                    "The data set to be merged does not give a complete isotropic exchange, please add more data"
                )
            Jiso[key] = Jiso_sum / weights
        self.dat.exchange_Jdict = Jiso

    def write(self, path="TB2J_results"):
        """
        Write the merged TB2J results to a folder.
        :param path: the path to the folder to write the results.
        """
        self.dat.description += (
            "Merged from TB2J results in paths: \n  " + "\n  ".join(self.paths) + "\n"
        )
        if self.method == "spin":
            self.dat.description += (
                ", which are from DFT data with various spin orientations. \n"
            )
        elif self.method == "structure":
            self.dat.description += ", which are from DFT data with structure with z axis rotated to x, y, z\n"
        self.dat.description += "\n"
        self.dat.write_all(path=path)


class Merger:
    def __init__(self, path_x, path_y, path_z, method="structure"):
        assert method in ["structure", "spin"]
        self.dat_x = read_pickle(path_x)
        self.dat_y = read_pickle(path_y)
        self.dat_z = read_pickle(path_z)
        self.dat = copy.copy(self.dat_z)
        self.paths = [path_x, path_y, path_z]
        self.method = method

    def merge_Jani(self):
        Jani_dict = {}
        Janixdict = self.dat_x.Jani_dict
        Janiydict = self.dat_y.Jani_dict
        Janizdict = self.dat_z.Jani_dict
        for key, Janiz in Janizdict.items():
            try:
                R, i, j = key
                keyx = R
                keyy = R
                Janix = Janixdict[(tuple(keyx), i, j)]
                Janiy = Janiydict[(tuple(keyy), i, j)]
            except KeyError as err:
                raise KeyError(
                    "Can not find key: %s, Please make sure the three calculations use the same k-mesh and same Rcut."
                    % err
                )
            if self.method == "spin":
                Jani_dict[key] = merge_Jani(Janix, Janiy, Janiz)
            else:
                Jani_dict[key] = merge_Jani(
                    swap_direction(Janix, (0, 2)), swap_direction(Janiy, (1, 2)), Janiz
                )
        self.dat.Jani_dict = Jani_dict

    def merge_Jiso(self):
        Jdict = {}
        Jxdict = self.dat_x.exchange_Jdict
        Jydict = self.dat_y.exchange_Jdict
        Jzdict = self.dat_z.exchange_Jdict
        for key, J in Jzdict.items():
            try:
                Jx = Jxdict[key]
                Jy = Jydict[key]
                Jz = Jzdict[key]
            except KeyError as err:
                raise KeyError(
                    "Can not find key: %s, Please make sure the three calculations use the same k-mesh and same Rcut."
                    % err
                )
            Jdict[key] = (Jx + Jy + Jz) / 3.0
        self.dat.exchange_Jdict = Jdict

    def merge_DMI(self):
        dmi_ddict = {}
        if self.dat_x.has_dmi and self.dat_y.has_dmi and self.dat_z.has_dmi:
            Dxdict = self.dat_x.dmi_ddict
            Dydict = self.dat_y.dmi_ddict
            Dzdict = self.dat_z.dmi_ddict
            for key, Dz in Dzdict.items():
                try:
                    R, i, j = key
                    keyx = R
                    keyy = R
                    Dx = Dxdict[(tuple(keyx), i, j)]
                    Dy = Dydict[(tuple(keyy), i, j)]
                except KeyError as err:
                    raise KeyError(
                        "Can not find key: %s, Please make sure the three calculations use the same k-mesh and same Rcut."
                        % err
                    )
                if self.method == "structure":
                    dmi_ddict[key] = rot_merge_DMI(Dx, Dy, Dz)
                else:
                    dmi_ddict[key] = merge_DMI(Dx, Dy, Dz)
            self.dat.dmi_ddict = dmi_ddict

        dmi_ddict = {}
        try:
            Dxdict = self.dat_x.debug_dict["DMI2"]
            Dydict = self.dat_y.debug_dict["DMI2"]
            Dzdict = self.dat_z.debug_dict["DMI2"]
            for key, Dz in Dzdict.items():
                try:
                    R, i, j = key
                    keyx = R
                    keyy = R
                    Dx = Dxdict[(tuple(keyx), i, j)]
                    Dy = Dydict[(tuple(keyy), i, j)]
                except KeyError as err:
                    raise KeyError(
                        "Can not find key: %s, Please make sure the three calculations use the same k-mesh and same Rcut."
                        % err
                    )
                if self.method == "structure":
                    dmi_ddict[key] = rot_merge_DMI2(Dx, Dy, Dz)
                elif self.method == "spin":
                    dmi_ddict[key] = merge_DMI2(Dx, Dy, Dz)

            self.dat.debug_dict["DMI2"] = dmi_ddict
        except:
            pass

    def write(self, path="TB2J_results"):
        self.dat.description += (
            "Merged from TB2J results in paths: \n  " + "\n  ".join(self.paths) + "\n"
        )
        if self.method == "spin":
            self.dat.description += (
                ", which are from DFT data with spin along x, y, z orientation\n"
            )
        elif self.method == "structure":
            self.dat.description += ", which are from DFT data with structure with z axis rotated to x, y, z\n"
        self.dat.description += "\n"
        self.dat.write_all(path=path)


def merge(path_x, path_y, path_z, method, save=True, path="TB2J_results"):
    m = Merger(path_x, path_y, path_z, method)
    m.merge_Jiso()
    m.merge_DMI()
    m.merge_Jani()
    if save:
        m.write(path=path)
    return m.dat


def merge2(paths, method, save=True, path="TB2J_results"):
    """
    Merge TB2J results from multiple calculations.
    :param paths: a list of paths to the TB2J results.
    :param method: 'structure' or 'spin'
    :param save: whether to save the merged results.
    :param path: the path to the folder to write the results.
    """
    m = Merger2(paths, method)
    m.merge_Jiso()
    m.merge_DMI()
    m.merge_Jani()
    if save:
        m.write(path=path)
    return m.dat



import numpy as np
import scipy.linalg as sl


def eigen_to_G(evals, evecs, efermi, energy):
    """calculate green's function from eigenvalue/eigenvector for energy(e-ef): G(e-ef).
    :param evals:  eigen values
    :param evecs:  eigen vectors
    :param efermi: fermi energy
    :param energy: energy
    :returns: Green's function G,
    :rtype:  Matrix with same shape of the Hamiltonian (and eigenvector)
    """
    return evecs.dot(np.diag(1.0 / (-evals + energy + efermi))).dot(evecs.conj().T)


def H2G(H, energy):
    return np.linalg.inv(energy * np.eye(len(H)) - H)


def Gpert(H0, dH, e):
    """
    Show Dyson
    """
    evals, evecs = sl.eigh(H0)
    G0 = eigen_to_G(evals, evecs, 0.0, e)

    G01 = H2G(H0, e)
    print(G0 - G01)

    evals, evecs = sl.eigh(H0 + dH)
    G = eigen_to_G(evals, evecs, 0.0, e)

    print(G - (G0 @ dH @ G + G0))
    print(G - (G0 @ dH @ G0 + G0))
    # print(G0@dH@G0+G0@dH@G0@dH@G0-G+G0)
    # print(G0@dH@G0+G0@dH@G0@dH@G0+G0@dH@G0@dH@G0@dH@G0-G+G0)


def test():
    H0 = np.random.rand(4, 4)
    H0 = H0 + H0.T.conj()

    dH = np.random.rand(4, 4)
    dH = H0 + H0.T.conj()

    dH = dH * 0.1

    Gpert(H0, dH, 1.3 + 0.00j)


# test()



import numpy as np
import os
from ase.units import J


def write_vampire(cls, path="TB2J_results/Vampire"):
    if not os.path.exists(path):
        os.makedirs(path)
    write_vampire_unitcell_file(cls, os.path.join(path, "vampire.UCF"))
    write_vampire_mat_file(cls, os.path.join(path, "vampire.mat"))
    write_vampire_inp_file(cls, os.path.join(path, "input"))


def write_vampire_unitcell_file(cls, fname):
    with open(fname, "w") as myfile:
        cell = cls.atoms.get_cell().array
        lattice_parameters = np.linalg.norm(cell, axis=-1)
        myfile.write("# Unit cell size (Angstrom):\n")
        np.savetxt(myfile, [lattice_parameters], fmt="%f")
        myfile.write("# Unit cell lattice vectors:\n")
        np.savetxt(myfile, cell / lattice_parameters, fmt="%f")

        myfile.write("# Atoms\n")
        nspins = sum([1 if i > -1 else 0 for i in cls.index_spin])
        myfile.write("%s %s\n" % (nspins, nspins))
        natom = len(cls.atoms)
        for i in range(natom):
            text = ""
            id_spin = cls.index_spin[i]
            if id_spin > -1:
                pos = cls.atoms.get_scaled_positions()[i]
                text = "{id_spin} {pos_x} {pos_y} {pos_z} {mat_id}\n".format(
                    id_spin=id_spin,
                    pos_x=pos[0],
                    pos_y=pos[1],
                    pos_z=pos[2],
                    mat_id=id_spin,
                )
                myfile.write(text)
        myfile.write("# Interactions\n")

        nexch = len(cls.exchange_Jdict.items())
        myfile.write(
            "{num_interactions} {type_exchange}\n".format(
                num_interactions=nexch, type_exchange="tensorial"
            )
        )

        counter = -1
        for key in cls.exchange_Jdict:
            R, ispin, jspin = key
            Jtensor = cls.get_J_tensor(ispin, jspin, R)
            counter += 1  # starts at 0
            myfile.write(f"{counter} {ispin} {jspin} {R[0]} {R[1]} {R[2]} ")
            for i in range(3):
                for j in range(3):
                    myfile.write(f"{Jtensor[i,j]*2.0/J} ")
            myfile.write("\n")


def write_vampire_mat_file(cls, fname):
    mat_tmpl = """#---------------------------------------------------
# Material {id} 
#---------------------------------------------------
material[{id}]:material-name={name}
material[{id}]:damping-constant={damping}
material[{id}]:atomic-spin-moment={ms} !muB
material[{id}]:uniaxial-anisotropy-constant={k1}
material[{id}]:material-element={name}
material[{id}]:initial-spin-direction = {spinat}
material[{id}]:uniaxial-anisotropy-direction = {k1dir}
#---------------------------------------------------
"""
    with open(fname, "w") as myfile:
        nspins = sum([1 if i > -1 else 0 for i in cls.index_spin])
        myfile.write("material:num-materials = %s\n" % (nspins))
        nspins = sum([1 if i > -1 else 0 for i in cls.index_spin])
        natom = len(cls.atoms)
        for i in range(natom):
            text = ""
            id_spin = cls.index_spin[i]
            if id_spin > -1:
                name = cls.atoms.get_chemical_symbols()[i]
                damping = 1.0
                ms = np.sqrt(np.sum(np.array(cls.spinat[i]) ** 2))
                spin = np.array(cls.spinat[i]) / ms
                spin_text = ",".join(map(str, spin))
                if cls.k1 is not None:
                    k1 = cls.k1[id_spin - 1]
                    k1dir = ", ".join(map(str, cls.k1dir[id_spin - 1]))
                else:
                    k1 = 0.0
                    k1dir = "0.0 , 0.0, 1.0"

                text = mat_tmpl.format(
                    id=id_spin + 1,
                    damping=damping,
                    name=name,
                    ms=ms,
                    k1=k1 / J,
                    k1dir=k1dir,
                    spinat=spin_text,
                )
                myfile.write(text)
        myfile.write("# Interactions\n")


def write_vampire_inp_file(cls, fname):
    text = """#------------------------------------------ 
# Creation attributes: 
#------------------------------------------ 
create:full
create:periodic-boundaries-x 
create:periodic-boundaries-y 
create:periodic-boundaries-z 
#------------------------------------------
material:file=vampire.mat
material:unit-cell-file = "vampire.UCF"
#------------------------------------------ 
# System Dimensions: 
#------------------------------------------ 
dimensions:unit-cell-size-x = {a}
dimensions:unit-cell-size-y = {b}
dimensions:unit-cell-size-z = {c}

dimensions:system-size-x = 15.0 !nm 
dimensions:system-size-y = 15.0 !nm 
dimensions:system-size-z = 15.0 !nm 
#------------------------------------------ 
# Simulation attributes: 
#------------------------------------------ 
sim:temperature=300 
sim:minimum-temperature=0 
sim:maximum-temperature=1000
sim:temperature-increment=25
sim:time-steps-increment=1 
sim:equilibration-time-steps=2500 
sim:loop-time-steps=3000 
#------------------------------------------ 
# Program and integrator details 
#------------------------------------------ 
sim:program=curie-temperature
sim:integrator=llg-heun
#------------------------------------------
# Data output 
#------------------------------------------ 
#output:real-time
output:temperature
output:material-magnetisation 
#output:magnetisation-length 
"""
    with open(fname, "w") as myfile:
        # cellpar = cls.atoms.get_cell_lengths_and_angles()
        cellpar = cls.atoms.cell.cellpar()
        myfile.write(text.format(a=cellpar[0], b=cellpar[1], c=cellpar[2]))



#!/usr/bin/env python
import math
import numpy as np
import scipy.linalg as linalg
from TB2J.kpoints import monkhorst_pack
from ase.dft.dos import DOS
import matplotlib.pyplot as plt
from ase.units import eV, J


class QSolver(object):
    def __init__(self, hamiltonian):
        self.ham = hamiltonian
        self.nspin = self.ham.nspin
        if len(self.ham.spin.shape) == 2:
            M = linalg.norm(self.ham.spin, axis=1)
        else:
            M = self.ham.spin

        self.M_mat = np.kron(np.sqrt(np.einsum("i,j->ij", M, M)), np.ones((3, 3)))
        self.Eref = None

    def get_Eref(self):
        mat = np.zeros((3 * self.nspin, 3 * self.nspin), dtype=complex)
        for key, val in self.ham.get_total_hessian_ijR().items():
            i, j, R = key
            mat[i * 3 : i * 3 + 3, j * 3 : j * 3 + 3] -= val
        mat = mat * 4.0 / self.M_mat
        SZ = np.zeros(self.nspin * 3, dtype=float)
        SZ[2::3] = 1.0
        SZ /= np.linalg.norm(SZ)
        self.Eref = np.dot(SZ, np.dot(mat, SZ)).real
        return self.Eref

    def solve_k(self, kpt, eigen_vectors=True, Jq=False):
        if self.Eref is None and not Jq:
            self.get_Eref()
        mat = np.zeros((3 * self.nspin, 3 * self.nspin), dtype=complex)
        for key, val in self.ham.get_total_hessian_ijR().items():
            i, j, R = key
            mat[i * 3 : i * 3 + 3, j * 3 : j * 3 + 3] -= val * np.exp(
                2.0j * math.pi * np.dot(kpt, R)
            )
        if not Jq:
            mat = mat * 4.0 / self.M_mat
        if eigen_vectors:
            evals, evecs = linalg.eigh(mat)
            evals = np.real(evals)
            if not Jq:
                evals -= self.Eref
            return evals, evecs
        else:
            evals = np.linalg.eigvalsh(mat)
            if not Jq:
                evals -= self.Eref
            return evals

    def solve_all(self, kpts, eigen_vectors=True, Jq=False):
        eval_list = []
        evec_list = []
        if eigen_vectors:
            for kpt in kpts:
                evals, evecs = self.solve_k(kpt, eigen_vectors=True, Jq=Jq)
                eval_list.append(evals)
                evec_list.append(evecs)
            return np.array(eval_list), np.array(evec_list)
        else:
            for kpt in kpts:
                evals = self.solve_k(kpt, eigen_vectors=False, Jq=Jq)
                eval_list.append(evals)
            return np.array(eval_list)


class QSolverASEWrapper(QSolver):
    def set(self, atoms=None, kmesh=[9, 9, 9], gamma=True, Jq=False, **kwargs):
        self.atoms = atoms
        self.dos_args = {
            "kmesh": kmesh,
            "gamma": gamma,
            "Jq": Jq,
        }
        self.dos_kwargs = kwargs
        self.kpts = monkhorst_pack(
            self.dos_args["kmesh"], gamma_center=self.dos_args["gamma"]
        )
        self.weights = np.ones(len(self.kpts)) / len(self.kpts)

    def get_k_points_and_weights(self):
        return self.kpts, self.weights

    def get_k_point_weights(self):
        return self.weights

    def get_number_of_spins(self):
        return 1

    def get_eigenvalues(self, kpt, spin=0):
        """
        return the eigenvalues at a given k-point. The energy unit is eV
        args:
            kpt: k-point index.
            spin: spin index.
        """
        kpoint = self.kpts[kpt]
        evals = self.solve_k(kpoint, eigen_vectors=False, Jq=self.dos_args["Jq"])
        evals = evals * J / eV
        return evals

    def get_fermi_level(self):
        return 0.0

    def get_bz_k_points(self):
        return self.kpts

    def get_dos(self, width=0.1, window=None, npts=401):
        dos = DOS(self, width=width, window=window, npts=npts)
        energies = dos.get_energies()
        tdos = dos.get_dos()
        return energies, tdos

    def plot_dos(
        self,
        smearing_width=0.0001,
        window=None,
        npts=401,
        output="dos.pdf",
        ax=None,
        show=True,
        dos_filename="magnon_dos.txt",
    ):
        """
        plot total DOS.
        :param width: width of Gaussian smearing
        :param window: energy window
        :param npts: number of points
        :param output: output filename
        :param ax: matplotlib axis
        :return: ax
        """
        if ax is None:
            _fig, ax = plt.subplots()
        energies, tdos = self.get_dos(width=smearing_width, window=window, npts=npts)
        energies = energies * 1000
        tdos = tdos / 1000
        if dos_filename is not None:
            np.savetxt(
                dos_filename,
                np.array([energies, tdos]).T,
                header="Energy(meV) DOS(state/meV)",
            )
        ax.plot(energies, tdos)
        ax.set_xlabel("Energy (meV)")
        ax.set_ylabel("DOS (states/meV)")
        ax.set_title("Total DOS")
        if output is not None:
            plt.savefig(output)
        if show:
            plt.show()
        return ax



TB2J_ref = """
% The implementation of TB2J:

@article{he2021tb2j,
title = {{TB2J}: A python package for computing magnetic interaction parameters},
journal = {Computer Physics Communications},
volume = {264},
pages = {107938},
year = {2021},
issn = {0010-4655},
doi = {https://doi.org/10.1016/j.cpc.2021.107938},
url = {https://www.sciencedirect.com/science/article/pii/S0010465521000679},
author = {Xu He and Nicole Helbig and Matthieu J. Verstraete and Eric Bousquet},
}


"""


LKAG_ref = """
% The original idea of the magnetic force theorem method: 

@article{LKAG,
title = {Local spin density functional approach to the theory of exchange interactions in ferromagnetic metals and alloys},
journal = {Journal of Magnetism and Magnetic Materials},
volume = {67},
number = {1},
pages = {65-74},
year = {1987},
issn = {0304-8853},
doi = {https://doi.org/10.1016/0304-8853(87)90721-9},
url = {https://www.sciencedirect.com/science/article/pii/0304885387907219},
author = {A.I. Liechtenstein and M.I. Katsnelson and V.P. Antropov and V.A. Gubanov},
}


"""


Wannier_LKAG_ref = """
% Uisng Wannier function Hamiltonian:

@article{PhysRevB.91.224405,
  title = {Calculation of exchange constants of the Heisenberg model in plane-wave-based methods using the Green's function approach},
  author = {Korotin, Dm. M. and Mazurenko, V. V. and Anisimov, V. I. and Streltsov, S. V.},
  journal = {Phys. Rev. B},
  volume = {91},
  issue = {22},
  pages = {224405},
  numpages = {7},
  year = {2015},
  month = {Jun},
  publisher = {American Physical Society},
  doi = {10.1103/PhysRevB.91.224405},
  url = {https://link.aps.org/doi/10.1103/PhysRevB.91.224405}
}


"""


NonCollinear_ref = """
% Extension to DMI and anisotropic exchange:

@article{Antropov1997,
  title={Exchange interactions in magnets},
  author={Antropov, VP and Katsnelson, MI and Liechtenstein, AI},
  journal={Physica B: Condensed Matter},
  volume={237},
  pages={336--340},
  year={1997},
  publisher={Elsevier}
}


"""


downfolding_ref = """
% Downfolding method to include the ligand contribution:

@article{PhysRevB.103.104428,
  title = {Exchange interactions and magnetic force theorem},
  author = {Solovyev, I. V.},
  journal = {Phys. Rev. B},
  volume = {103},
  issue = {10},
  pages = {104428},
  numpages = {14},
  year = {2021},
  month = {Mar},
  publisher = {American Physical Society},
  doi = {10.1103/PhysRevB.103.104428},
  url = {https://link.aps.org/doi/10.1103/PhysRevB.103.104428}
}


"""


def print_bib():
    print("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")
    print("%     We recommend to cite the following references:       %")
    print("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")
    print(TB2J_ref)

    print(LKAG_ref)
    print(Wannier_LKAG_ref)
    print(NonCollinear_ref)
    print(downfolding_ref)


if __name__ == "__main__":
    print_bib()



import numpy as np


def monkhorst_pack(size, gamma_center=False):
    """Construct a uniform sampling of k-space of given size.
    Modified from ase.dft.kpoints with gamma_center option added"""
    if np.less_equal(size, 0).any():
        raise ValueError("Illegal size: %s" % list(size))
    kpts = np.indices(size).transpose((1, 2, 3, 0)).reshape((-1, 3))
    asize = np.array(size)
    shift = 0.5 * ((asize + 1) % 2) / asize
    mkpts = (kpts + 0.5) / size - 0.5
    if gamma_center:
        mkpts += shift

    return mkpts



#!/usr/bin/env python3
from TB2J.plot import plot_magnon_band, write_eigen
from TB2J.versioninfo import print_license
import argparse

"""
The script to plot the magnon band structure.
"""


def plot_magnon():
    print_license()
    parser = argparse.ArgumentParser(
        description="TB2J_magnon: Plot magnon band structure from the TB2J magnetic interaction parameters"
    )
    parser.add_argument(
        "--fname",
        default="exchange.xml",
        type=str,
        help="exchange xml file name. default: exchange.xml",
    )

    parser.add_argument(
        "--qpath",
        default=None,
        type=str,
        help="The names of special q-points. If not given, the path will be automatically choosen. See https://wiki.fysik.dtu.dk/ase/ase/dft/kpoints.html for the table of special kpoints and the default path.",
    )

    parser.add_argument(
        "--figfname",
        default=None,
        type=str,
        help="The file name of the figure. It should be e.g. png, pdf or other types of files which could be generated by matplotlib.",
    )

    parser.add_argument(
        "--Jq",
        action="store_true",
        help="To plot the eigenvalues of -J(q) instead of the magnon band",
        default=False,
    )

    parser.add_argument(
        "--no_Jiso",
        action="store_true",
        help="switch off isotropic exchange",
        default=None,
    )

    parser.add_argument(
        "--no_dmi", action="store_true", help="switch off dmi", default=None
    )

    parser.add_argument(
        "--no_Jani",
        action="store_true",
        help="switch off anisotropic exchange",
        default=None,
    )

    parser.add_argument(
        "--write_emin", action="store_true", help="switch to write emin", default=False
    )

    parser.add_argument(
        "--show",
        action="store_true",
        help="whether to show magnon band structure.",
        default=False,
    )

    args = parser.parse_args()
    if args.Jq:
        if args.figfname is None:
            args.figfname = "Eigen_Jq.pdf"
        print(
            "Plotting the eigenvalues of -J(q). The figure is written to %s"
            % (args.figfname)
        )
    else:
        if args.figfname is None:
            args.figfname = "magnon_band.pdf"
        print(
            "Plotting the magnon band structure. The figure is written to %s"
            % (args.figfname)
        )

    def nonone(x):
        if x is None:
            return x
        else:
            return not x

    if args.write_emin:
        write_eigen(
            has_exchange=nonone(args.no_Jiso),
            has_dmi=nonone(args.no_dmi),
            has_bilinear=nonone(args.no_Jani),
        )
    else:
        plot_magnon_band(
            fname=args.fname,
            knames=args.qpath,
            npoints=300,
            Jq=args.Jq,
            figfname=args.figfname,
            show=args.show,
            has_exchange=nonone(args.no_Jiso),
            has_dmi=nonone(args.no_dmi),
            has_bilinear=nonone(args.no_Jani),
        )


if __name__ == "__main__":
    plot_magnon()



#!/usr/bin/env python3
import argparse
import sys
from TB2J.manager import gen_exchange
from TB2J.versioninfo import print_license


def run_wann2J():
    print_license()
    parser = argparse.ArgumentParser(
        description="wann2J: Using magnetic force theorem to calculate exchange parameter J from wannier functions"
    )
    parser.add_argument(
        "--path", help="path to the wannier files", default="./", type=str
    )
    parser.add_argument(
        "--posfile", help="name of the position file", default="POSCAR", type=str
    )
    parser.add_argument(
        "--prefix_spinor",
        help="prefix to the spinor wannier files",
        default="wannier90",
        type=str,
    )
    parser.add_argument(
        "--prefix_up",
        help="prefix to the spin up wannier files",
        default="wannier90.up",
        type=str,
    )
    parser.add_argument(
        "--prefix_down",
        help="prefix to the spin down wannier files",
        default="wannier90.dn",
        type=str,
    )
    parser.add_argument(
        "--elements",
        help="elements to be considered in Heisenberg model",
        default=None,
        type=str,
        nargs="*",
    )
    parser.add_argument(
        "--groupby",
        help="In the spinor case, the order of the orbitals have two conventions: 1: group by spin (orb1_up, orb2_up,... orb1_down, ...), 2,group by orbital (orb1_up, orb1_down, orb2_up, ...,). Use 'spin' in the former case and 'orbital' in the latter case. The default is spin.",
        default="spin",
        type=str,
    )
    parser.add_argument(
        "--write_dm",
        help="whether to write density matrix",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--rcut",
        help="cutoff of spin pair distance. The default is to calculate all commensurate R point to the k mesh.",
        default=None,
        type=float,
    )
    parser.add_argument("--efermi", help="Fermi energy in eV", default=None, type=float)
    parser.add_argument(
        "--kmesh",
        help="kmesh in the format of kx ky kz",
        type=int,
        nargs="*",
        default=[5, 5, 5],
    )
    parser.add_argument(
        "--emin",
        help="energy minimum below efermi, default -14 eV",
        type=float,
        default=-14.0,
    )
    parser.add_argument(
        "--emax",
        help="energy maximum above efermi, default 0.0 eV",
        type=float,
        default=0.0,
    )
    parser.add_argument(
        "--nz",
        help="number of steps for semicircle contour, default: 100",
        default=100,
        type=int,
    )
    parser.add_argument(
        "--cutoff",
        help="The minimum of J amplitude to write, (in eV), default is 1e-5 eV",
        default=1e-5,
        type=float,
    )
    parser.add_argument(
        "--exclude_orbs",
        help="the indices of wannier functions to be excluded from magnetic site. counting start from 0",
        default=[],
        type=int,
        nargs="+",
    )

    parser.add_argument(
        "--np",
        help="number of cpu cores to use in parallel, default: 1",
        default=1,
        type=int,
    )

    parser.add_argument(
        "--use_cache",
        help="whether to use disk file for temporary storing wavefunctions and hamiltonian to reduce memory usage. Default: False",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--description",
        help="add description of the calculatiion to the xml file. Essential information, like the xc functional, U values, magnetic state should be given.",
        type=str,
        default="Calculated with TB2J.",
    )

    parser.add_argument(
        "--spinor",
        action="store_true",
        help="Whether to use spinor wannier function.",
        default=False,
    )

    parser.add_argument(
        "--orb_decomposition",
        default=False,
        action="store_true",
        help="whether to do orbital decomposition in the non-collinear mode.",
    )

    parser.add_argument(
        "--output_path",
        help="The path of the output directory, default is TB2J_results",
        type=str,
        default="TB2J_results",
    )

    parser.add_argument(
        "--wannier_type",
        help="The type of Wannier function, either Wannier90 or banddownfolder",
        type=str,
        default="Wannier90",
    )

    # parser.add_argument("--qspace",
    #                    action="store_true",
    #                    help="Whether to calculate J in qspace first and transform to real space.",
    #                    default=False)

    args = parser.parse_args()

    if args.efermi is None:
        print("Please input fermi energy using --efermi ")
        sys.exit()
    if args.elements is None:
        print("Please input the magnetic elements, e.g. --elements Fe Ni")
        sys.exit()

    gen_exchange(
        path=args.path,
        colinear=(not args.spinor),
        groupby=args.groupby,
        posfile=args.posfile,
        efermi=args.efermi,
        kmesh=args.kmesh,
        magnetic_elements=args.elements,
        Rcut=args.rcut,
        prefix_SOC=args.prefix_spinor,
        prefix_up=args.prefix_up,
        prefix_dn=args.prefix_down,
        emin=args.emin,
        emax=args.emax,
        nz=args.nz,
        use_cache=args.use_cache,
        np=args.np,
        description=args.description,
        output_path=args.output_path,
        exclude_orbs=args.exclude_orbs,
        wannier_type=args.wannier_type,
        # qspace=args.qspace,
        write_density_matrix=args.write_dm,
        orb_decomposition=args.orb_decomposition,
    )


if __name__ == "__main__":
    run_wann2J()
